# coding: utf-8
"""This module contains the class defining Uniform 3D meshes."""
from __future__ import annotations

import numpy as np

#from itertools import product as iproduct
from monty.functools import lazy_property
#from numpy.random import random
from numpy.fft import fftn, ifftn, fftshift, ifftshift, fftfreq
from abipy.tools import duck


__all__ = [
    "Mesh3D",
]


class Mesh3D:
    r"""
    Descriptor-class for uniform 3D meshes.

    A Mesh3D object holds information on how functions, such
    as wave functions and electron densities, are discretized in a
    certain domain in space. The main information here is how many
    grid points are used in each direction of the unit cell.

    There are methods for tasks such as allocating arrays, performing
    symmetry operations and integrating functions over space.

    This is how a 2x2x2 3D array arr[x,y,z] is stored in memory::

        3-----7
        |\    |
        | \   |
        |  1-----5      z
        2--|--6  |   y  |
         \ |   \ |    \ |
          \|    \|     \|
           0-----4      +-----x

    """
    def __init__(self, shape, vectors):
        """
        Construct ``Mesh3D`` object.

        Args:
            shape: 3 int's Number of grid points along axes.
            vectors: unit cell vectors in real space.

        Attributes:

        ==========  ========================================================
        ``shape``   Array of the number of grid points along the three axes.
        ``dv``      Volume per grid point.
        ==========  ========================================================
        """
        self.shape = tuple(np.asarray(shape, int))
        self.size = np.prod(self.shape)
        self.vectors = np.reshape(vectors, (3, 3))

        cross12 = np.cross(self.vectors[1], self.vectors[2])
        self.dv = abs(np.sum(self.vectors[0] * cross12.T)) / self.size
        self.dvx = self.vectors[0] / self.nx
        self.dvy = self.vectors[1] / self.ny
        self.dvz = self.vectors[2] / self.nz

    def __len__(self):
        return self.size

    def __eq__(self, other):
        return (np.all(self.shape == other.shape) and
                np.all(self.vectors == other.vectors))

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        return self.to_string()

    def __iter__(self):
        for ix in range(self.nx):
            for iy in range(self.ny):
                for iz in range(self.nz):
                    yield ix * self.dvx + iy * self.dvy + iz * self.dvz

    def iter_ixyz_r(self):
        """
        Iterator returning (ixyz, rr) where ixyz gives the index of the point and r is the point on the grid.
        """
        for ix in range(self.nx):
            for iy in range(self.ny):
                for iz in range(self.nz):
                    rr = ix * self.dvx + iy * self.dvy + iz * self.dvz
                    yield np.array((ix, iy, iz), dtype=int), rr

    def rpoint(self, ix, iy, iz):
        """The vector corresponding to the (ix, iy, iz) indices"""
        return ix * self.dvx + iy * self.dvy + iz * self.dvz

    def to_string(self, verbose=0) -> str:
        """String representation."""
        return self.__class__.__name__ + ": nx=%d, ny=%d, nz=%d" % self.shape

    @property
    def nx(self) -> int:
        """Number of points along x."""
        return self.shape[0]

    @property
    def ny(self) -> int:
        """Number of points along y."""
        return self.shape[1]

    @property
    def nz(self) -> int:
        """Number of points along z."""
        return self.shape[2]

    @lazy_property
    def inv_vectors(self):
        return np.linalg.inv(self.vectors)

    def _new_array(self, dtype=float, zero=True, extra_dims=()):
        shape = self.shape

        if duck.is_intlike(extra_dims):
            extra_dims = (extra_dims,)

        shape = extra_dims + tuple(shape)

        if zero:
            return np.zeros(shape, dtype)
        else:
            return np.empty(shape, dtype)

    def zeros(self, dtype=float, extra_dims=()) -> np.ndarray:
        """
        Returns new zeroed 3D array for this domain.

        The type can be set with the ``dtype`` keyword.
        Extra dimensions can be added with ``extra_dims``.
        """
        return self._new_array(dtype=dtype, zero=True, extra_dims=extra_dims)

    def czeros(self, extra_dims=()) -> np.ndarray:
        """Returns new zeroed 3D complex array for this domain."""
        return self._new_array(dtype=complex, zero=True, extra_dims=extra_dims)

    def empty(self, dtype=float, extra_dims=()):
        """
        Returns new uninitialized 3D |numpy-array| for this domain.

        The type can be set with the ``dtype`` keyword.
        Extra dimensions can be added with ``extra_dims``.
        """
        return self._new_array(dtype=dtype, zero=False, extra_dims=extra_dims)

    def cempty(self, extra_dims=()) -> np.ndarray:
        """Returns new uninitialized 3D complex |numpy-array| for this domain."""
        return self._new_array(dtype=complex, zero=False, extra_dims=extra_dims)

    def random(self, dtype=float, extra_dims=()) -> np.ndarray:
        """Returns random real |numpy-array| for this domain with val in [0.0, 1.0)."""
        shape = self.shape
        if duck.is_intlike(extra_dims):
            extra_dims = (extra_dims,)

        shape = extra_dims + tuple(shape)

        re = np.random.random(shape)
        if dtype == float:
            return re
        elif dtype == complex:
            im = self.random(extra_dims=extra_dims)
            return re + 1j*im
        else:
            raise ValueError("Wrong dtype: %s" % str(dtype))

    def crandom(self, extra_dims=()) -> np.ndarray:
        """Returns random complex |numpy-array| for this domain with val in [0.0, 1.0)."""
        return self.random(dtype=complex, extra_dims=extra_dims)

    def reshape(self, arr) -> np.ndarray:
        """
        Reshape the array arr defined on the FFT box.

        Returns |numpy-array| with 4 dimensions (?, nx, ny, nz) where ?*nx*ny*nz == arr.size
        """
        #if duck.is_intlike(extra_dims): extra_dims = (extra_dims,)
        #shape = extra_dims + self.shape)
        return np.reshape(arr, (-1,) + self.shape)

    def fft_r2g(self, fr, shift_fg=False) -> np.ndarray:
        """
        FFT of array ``fr`` given in real space.
        """
        ndim, shape = fr.ndim, fr.shape

        if ndim == 1:
            fr = np.reshape(fr, self.shape)
            return self.fft_r2g(fr, shift_fg=shift_fg).flatten()

        elif ndim == 3:
            assert self.size == np.prod(shape[-3:])
            fg = fftn(fr)
            if shift_fg: fg = fftshift(fg)

        elif ndim > 3:
            assert self.size == np.prod(shape[-3:])
            axes = np.arange(ndim)[-3:]
            fg = fftn(fr, axes=axes)
            if shift_fg: fg = fftshift(fg, axes=axes)

        else:
            raise NotImplementedError("ndim < 3 are not supported")

        return fg / self.size

    def fft_g2r(self, fg, fg_ishifted=False) -> np.ndarray:
        """
        FFT of array ``fg`` given in G-space.
        """
        ndim, shape = fg.ndim, fg.shape

        if ndim == 1:
            fg = np.reshape(fg, self.shape)
            return self.fft_g2r(fg, fg_ishifted=fg_ishifted).flatten()

        if ndim == 3:
            assert self.size == np.prod(shape[-3:])
            if fg_ishifted: fg = ifftshift(fg)
            fr = ifftn(fg)

        elif ndim > 3:
            assert self.size == np.prod(shape[-3:])
            axes = np.arange(ndim)[-3:]
            if fg_ishifted: fg = ifftshift(fg, axes=axes)
            fr = ifftn(fg, axes=axes)

        else:
            raise NotImplementedError("ndim < 3 are not supported")

        return fr * self.size

    #def fourier_interp(self, data, new_mesh, inspace="r"):
    #    """
    #    Fourier interpolation of data.

    #    Args:

    #        data: Input array defined on this mesh
    #        new_mesh: Mesh where data is interpolated
    #        inspace: string specifying if data is given in real space "r" or in reciprocal space "g".

    #    Return:
    #        Numpy array in real space on the new_mesh
    #    """
    #    raise NotImplementedError("gtransfer is missing")
    #    assert inspace in ("r", "g")

    #    # Insert data in the FFT box of new mesh.
    #    if inspace == "r": data = self.fft_r2g(data)
    #    intp_datag = new_mesh.gtransfer_from(self, data)

    #    # FFT transform G --> R.
    #    return new_mesh.fft_g2r(intp_datag)

    def integrate(self, fr):
        """
        Integrate array(s) fr.
        """
        shape, ndim = fr.shape, fr.ndim
        assert self.size == np.prod(shape[-3:])

        if ndim == 3:
            return fr.sum() * self.dv

        elif ndim > 3:
            sums = np.sum(np.reshape(fr, shape[:-3] + (-1,)), axis=-1)
            return sums * self.dv

        else:
            raise NotImplementedError("ndim < 3 are not supported")

    @lazy_property
    def gvecs(self) -> np.ndarray:
        """
        Array with the reduced coordinates of the G-vectors.

        .. note::

            These are the G-vectors of the FFT box and should be used
            when we compute quantities on the FFT mesh.
            These vectors differ from the gvecs stored in |GSphere| that
            are k-centered and enclosed by a sphere whose radius is defined by ecut.
        """
        gx_list = np.rint(fftfreq(self.nx) * self.nx)
        gy_list = np.rint(fftfreq(self.ny) * self.ny)
        gz_list = np.rint(fftfreq(self.nz) * self.nz)
        #print(gz_list, gy_list, gx_list)

        gvecs = np.empty((self.size, 3), dtype=int)

        idx = -1
        for gx in gx_list:
            for gy in gy_list:
                for gz in gz_list:
                    idx += 1
                    gvecs[idx, :] = gx, gy, gz

        return gvecs

    @lazy_property
    def gmods(self) -> np.ndarray:
        """[ng] |numpy-array| with :math:`|G|`"""
        gmet = np.dot(self.inv_vectors.T, self.inv_vectors)
        gmods = np.empty(self.size)
        for i, g in enumerate(self.gvecs):
            gmods[i] = np.dot(g, np.dot(gmet, g))

        return 2 * np.pi * np.sqrt(gmods)

    #@lazy_property
    #def gmax(self)
    #    return self.gmods.max()

    @lazy_property
    def rpoints(self) -> np.ndarray:
        """|numpy-array| with the points in real space in reduced coordinates."""
        nx, ny, nz = self.nx, self.ny, self.nz
        rpoints = np.empty((self.size, 3))

        grid_points = np.meshgrid(np.linspace(0, 1, nx, endpoint=False),
                                  np.linspace(0, 1, ny, endpoint=False),
                                  np.linspace(0, 1, nz, endpoint=False))

        rpoints[:, 0] = grid_points[0].ravel()
        rpoints[:, 1] = grid_points[1].ravel()
        rpoints[:, 2] = grid_points[2].ravel()

        return rpoints

    #def ogrid_rfft(self):
    #    return np.ogrid[0:1:1/self.nx,
    #                    0:1:1/self.ny,
    #                    0:1:1/self.nz]

    #def line_inds(self, line):
    #    """
    #    Returns an ogrid with the indices associated to the specified line.

    #    Args:
    #        line:
    #            String specifying the type of line (in reduced coordinates),
    #            e.g. "x", "y", "z".
    #    """
    #    line = line.lower()

    #    if line == "x":
    #        return np.ogrid[0:self.nx, 0:1, 0:1]
    #    elif line == "y":
    #        return np.ogrid[0:1, 0:self.ny, 0:1]
    #    elif line == "z":
    #        return np.ogrid[0:1, 0:1, 0:self.nz]
    #    else:
    #        raise ValueError("Wrong line %s" % line)

    #def plane_inds(self, plane, h):
    #    """
    #    Returns an ogrid with the indices associated to the specified plane.

    #    Args:
    #        plane:
    #            String specifying the type of plane (in reduced coordinates),
    #            e.g. "xy" "xz" ...
    #        h:
    #            Index giving the position of the plane along the perpendicular.
    #    """
    #    plane = plane.lower()
    #    nx, ny, nz = self.nx, self.ny, self.nz

    #    if plane in ("xy", "yx"):
    #        return np.ogrid[0:nx, 0:ny, h:h+1]
    #    elif plane in ("xz", "zx"):
    #        return np.ogrid[0:nx, h:h+1, 0:nz]
    #    elif plane in ("yz", "zy"):
    #        return np.ogrid[h:h+1, 0:ny, 0:nz]
    #    else:
    #        raise ValueError("Wrong plane %s" % plane)

    #def irottable(self, symmops):
    #    nsym = len(symmops)
    #    nx, ny, nz = self.nx, self.ny, self.nz

    #    red2fft = np.diag([nx, ny, nz])
    #    fft2red = np.diag([1/nx, 1/ny, 1/nz])

    #    # For a fully compatible mesh, each mat in rotsm1_fft should be integer
    #    rotsm1_fft, tnons_fft = np.empty((nsym, 3, 3)), np.empty((nsym, 3))

    #    for isym, symmop in enumerate(symmops):
    #        rotm1_r, tau = symmop.rotm1_r, symmop.tau
    #        rotsm1_fft[isym] = np.dot(np.dot(red2fft, rotm1_r), fft2red)
    #        tnons_fft[isym] = np.dot(red2fft, tau)

    #    # Indeces of $R^{-1}(r-\tau)$ in the FFT box.
    #    irottable = np.empty((nsym, nx*ny*nz), dtype=int)

    #    #max_err = 0.0
    #    nxyz = np.array((nx, ny, nz), np.int)
    #    for isym in range(nsym):
    #        rm1_fft = rotsm1_fft[isym]
    #        tau_fft = tnons_fft[isym]
    #        for ifft, p1_fft in enumerate(iproduct(range(nx), range(ny), range(nz))):
    #            # Form R^-1 (r-\tau) in the FFT basis.
    #            p1_fft = np.array(p1_fft)
    #            prot_fft = np.dot(rm1_fft, p1_fft - tau_fft)
    #            #err = ABS(prot_fft - (ix, iy, iz)) / (nx, ny, nz)
    #            prot_fft = np.round(prot_fft)
    #            jx, jy, jz = prot_fft % nxyz
    #            irottable[isym, ifft] = jz + (jy * nz) + (jx * nx * ny)

    #    # Test
    #    #for isym in range(nsym):
    #    #    irottable[isym, ifft]
    #    #    rm1_fft = rotsm1_fft[isym]
    #    #    tau_fft = tnons_fft[isym]
    #    #    for ifft, p1_fft in enumerate(itertools.product(range(nx), range(ny), range(nz))):
    #    #        irot_fft == irottable[isym, ifft]

    #    return irottable

    def i_closest_gridpoints(self, points) -> np.ndarray:
        """
        Given a list of points, this function return a |numpy-array| with the indices of the closest gridpoint.
        """
        points = np.reshape(points, (-1, 3))
        inv_vectors = self.inv_vectors
        fcoords = [np.dot(point, inv_vectors) for point in points]
        inds = [[np.mod(int(np.rint(pc[ii]*self.shape[ii])), self.nx) for ii in range(3)] for pc in fcoords]
        return np.array(inds)

        # return [(int(np.rint(pc[ii]*self.nx)), int(np.rint(pc[0]*self.nx)),int(np.rint(pc[0]*self.nx))) for pc in coords]
        # ix = int(np.rint(coords[0]*self.nx))
        # iy = int(np.rint(coords[1]*self.ny))
        # iz = int(np.rint(coords[2]*self.nz))
        # return (ix, iy, iz)

    # @DW TODO Add test.
    def dist_gridpoints_in_spheres(self, points, radius):
        # c_ab = np.cross(self.vectors[0], self.vectors[1])
        # c_bc = np.cross(self.vectors[1], self.vectors[2])
        # c_ca = np.cross(self.vectors[2], self.vectors[0])
        # h_ab = np.abs(np.dot(c_ab, self.vectors[2]) / np.linalg.norm(c_ab))
        # h_bc = np.abs(np.dot(c_bc, self.vectors[0]) / np.linalg.norm(c_bc))
        # h_ca = np.abs(np.dot(c_ca, self.vectors[1]) / np.linalg.norm(c_ca))
        maxdiag = max([np.linalg.norm(self.dvx+self.dvy+self.dvz),
                       np.linalg.norm(self.dvx+self.dvy-self.dvz),
                       np.linalg.norm(self.dvx-self.dvy+self.dvz),
                       np.linalg.norm(self.dvx-self.dvy-self.dvz)])
        c_ab = np.cross(self.dvx, self.dvy)
        c_bc = np.cross(self.dvy, self.dvz)
        c_ca = np.cross(self.dvz, self.dvx)
        h_ab = np.abs(np.dot(c_ab, self.dvz) / np.linalg.norm(c_ab))
        h_bc = np.abs(np.dot(c_bc, self.dvx) / np.linalg.norm(c_bc))
        h_ca = np.abs(np.dot(c_ca, self.dvy) / np.linalg.norm(c_ca))
        a_factor = 1.01 * (radius+0.5*maxdiag) / h_bc
        b_factor = 1.01 * (radius+0.5*maxdiag) / h_ca
        c_factor = 1.01 * (radius+0.5*maxdiag) / h_ab
        # print('HEIGHTS')
        # print(h_ab, h_bc, h_ca)
        # print(c_factor*h_ab, a_factor* h_bc, b_factor*h_ca)
        mins = np.array(np.floor([-a_factor, -b_factor, -c_factor]), dtype=int)
        maxes = np.array(np.ceil([a_factor, b_factor, c_factor]), dtype=int)
        # print('AAAMINS AND MAXES', mins, maxes)
        # maxes = np.ceil(nmax)
        i_closest_gridpoint_points = self.i_closest_gridpoints(points=points)
        dist_gridpoints_points = []
        r2 = radius**2
        for ipoint, point_i_closest_gridpoint in enumerate(i_closest_gridpoint_points):
            pp = points[ipoint]
            dist_gridpoints = []
            for ix in range(int(mins[0]), int(maxes[0])):
                ipx = point_i_closest_gridpoint[0] + ix
                for iy in range(int(mins[1]), int(maxes[1])):
                    ipy = point_i_closest_gridpoint[1] + iy
                    for iz in range(int(mins[2]), int(maxes[2])):
                        ipz = point_i_closest_gridpoint[2] + iz
                        gp = ipx*self.dvx + ipy * self.dvy + ipz * self.dvz
                        dist2_gp_pp = np.dot(pp-gp, pp-gp)
                        if dist2_gp_pp <= r2:
                            dist_gridpoints.append(((np.mod(ipx, self.nx), np.mod(ipy, self.ny), np.mod(ipz, self.nz)),
                                                    np.sqrt(dist2_gp_pp), (ipx, ipy, ipz)))
            dist_gridpoints_points.append(dist_gridpoints)
        return dist_gridpoints_points

    # def dist2_gridpoints_in_spheres(self, points, radius):
    #     # c_ab = np.cross(self.vectors[0], self.vectors[1])
    #     # c_bc = np.cross(self.vectors[1], self.vectors[2])
    #     # c_ca = np.cross(self.vectors[2], self.vectors[0])
    #     # h_ab = np.abs(np.dot(c_ab, self.vectors[2]) / np.linalg.norm(c_ab))
    #     # h_bc = np.abs(np.dot(c_bc, self.vectors[0]) / np.linalg.norm(c_bc))
    #     # h_ca = np.abs(np.dot(c_ca, self.vectors[1]) / np.linalg.norm(c_ca))
    #     maxdiag = max([np.linalg.norm(self.dvx+self.dvy+self.dvz),
    #                    np.linalg.norm(self.dvx+self.dvy-self.dvz),
    #                    np.linalg.norm(self.dvx-self.dvy+self.dvz),
    #                    np.linalg.norm(self.dvx-self.dvy-self.dvz)])
    #     c_ab = np.cross(self.dvx, self.dvy)
    #     c_bc = np.cross(self.dvy, self.dvz)
    #     c_ca = np.cross(self.dvz, self.dvx)
    #     h_ab = np.abs(np.dot(c_ab, self.dvz) / np.linalg.norm(c_ab))
    #     h_bc = np.abs(np.dot(c_bc, self.dvx) / np.linalg.norm(c_bc))
    #     h_ca = np.abs(np.dot(c_ca, self.dvy) / np.linalg.norm(c_ca))
    #     a_factor = 1.01 * (radius+0.5*maxdiag) / h_bc
    #     b_factor = 1.01 * (radius+0.5*maxdiag) / h_ca
    #     c_factor = 1.01 * (radius+0.5*maxdiag) / h_ab
    #     # print('HEIGHTS')
    #     # print(h_ab, h_bc, h_ca)
    #     # print(c_factor*h_ab, a_factor* h_bc, b_factor*h_ca)
    #     mins = np.array(np.floor([-a_factor, -b_factor, -c_factor]), dtype=int)
    #     maxes = np.array(np.ceil([a_factor, b_factor, c_factor]), dtype=int)
    #     # print('AAAMINS AND MAXES', mins, maxes)
    #     # maxes = np.ceil(nmax)
    #     i_closest_gridpoint_points = self.i_closest_gridpoints(points=points)
    #     dist_gridpoints_points = []
    #     r2 = radius**2
    #     for ipoint, point_i_closest_gridpoint in enumerate(i_closest_gridpoint_points):
    #         pp = points[ipoint]
    #         dist_gridpoints = []
    #         for ix in range(int(mins[0]), int(maxes[0])):
    #             ipx = point_i_closest_gridpoint[0] + ix
    #             for iy in range(int(mins[1]), int(maxes[1])):
    #                 ipy = point_i_closest_gridpoint[1] + iy
    #                 for iz in range(int(mins[2]), int(maxes[2])):
    #                     ipz = point_i_closest_gridpoint[2] + iz
    #                     gp = ipx*self.dvx + ipy * self.dvy + ipz * self.dvz
    #                     dist2_gp_pp = np.linalg.norm(pp-gp)
    #                     if dist_gp_pp <= r2:
    #                         dist_gridpoints.append(((np.mod(ipx, self.nx), np.mod(ipy, self.ny), np.mod(ipz, self.nz)),
    #                                                 dist_gp_pp, (ipx, ipy, ipz)))
    #         dist_gridpoints_points.append(dist_gridpoints)
    #     return dist_gridpoints_points



#!/usr/bin/env python
r"""
Flow to compute e-ph scattering potentials
==========================================

This example shows how to compute e-ph scattering potentials
along a q-path, merge the POT files in the DVDB file and use the
DVDB and the DDB file to analyze the average over the unit cell of the
periodic part as a function of q
"""
import sys
import os
import abipy.abilab as abilab
import abipy.data as abidata

from abipy import flowtk


def make_scf_input(ngkpt):
    """
    This function constructs the input file for the GS calculation:
    """
    structure = abilab.Structure.from_abistring(
"""
acell 1.0522E+01  1.0522E+01  1.0522E+01
rprim  0.0  0.5 0.5
       0.5  0.0 0.5
       0.5  0.5 0.0

ntypat 1
znucl  32
natom 2
typat 1 1
xred  0.0  0.0  0.0
      1/4  1/4  1/4
""")

    pseudos = abidata.pseudos("Ge.psp8")
    gs_inp = abilab.AbinitInput(structure, pseudos=pseudos)

    gs_inp.set_vars(
        nband=16,
        ecut=40.0,
        ngkpt=ngkpt,
        nshiftk=1,
        shiftk=[0, 0, 0],
        tolvrs=1.0e-8,
        nstep=150,
        paral_kgb=0,
    )

    return gs_inp


def build_flow(options):
    """
    Create a `Flow` for phonon calculations. The flow has two works.

    The first work contains a single GS task that produces the WFK file used in DFPT
    Then we have multiple Works that are generated automatically
    in order to compute the dynamical matrix on a [2, 2, 2] mesh.
    Symmetries are taken into account: only q-points in the IBZ are generated and
    for each q-point only the independent atomic perturbations are computed.
    """
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_", "flow_")

    # Use 2x2x2 both for k-mesh and q-mesh
    # Build input for GS calculation
    scf_input = make_scf_input(ngkpt=(8, 8, 8))

    # Create flow to compute all the independent atomic perturbations
    # corresponding to a [4, 4, 4] q-mesh.
    # Electric field and Born effective charges are also computed.
    from abipy.flowtk.eph_flows import EphPotFlow
    ngqpt = [4, 4, 4]

    qpath_list = [
        [+0.000, +0.000, +0.000],  # name: $\Gamma$, weight: 0.000
        [+0.500, +0.000, +0.500],  # name: X, weight: 0.000
        [+0.500, +0.250, +0.750],  # name: W, weight: 0.000
        [+0.375, +0.375, +0.750],  # name: K, weight: 0.000
        [+0.000, +0.000, +0.000],  # name: $\Gamma$, weight: 0.000
        [+0.500, +0.500, +0.500],  # name: L, weight: 0.000
        #[+0.625, +0.250, +0.625],  # name: U, weight: 0.000
        #[+0.500, +0.250, +0.750],  # name: W, weight: 0.000
        #[+0.500, +0.500, +0.500],  # name: L, weight: 0.000
        #[+0.375, +0.375, +0.750],  # name: K, weight: 0.000
        #[+0.625, +0.250, +0.625],  # name: U, weight: 0.000
        #[+0.500, +0.000, +0.500],  # name: X, weight: 0.000
    ]

    # Use small ndivsm to reduce computing time.
    flow = EphPotFlow.from_scf_input(options.workdir, scf_input,
                                     ngqpt, qpath_list, ndivsm=10,
                                     with_quads=True,
                                     with_becs=True)

    return flow


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
r"""
Density
=======

This example shows how to analyze the electronic density
stored in the DEN.nc file.
"""
from abipy.abilab import abiopen
import abipy.data as abidata

# Open the DEN.nc file
ncfile = abiopen(abidata.ref_file("si_DEN.nc"))

# The DEN file has a `Density`, a `Structure` and an `ElectronBands` object
print(ncfile.structure)

#%%
# To plot the KS eigenvalues.
ncfile.ebands.plot()

density = ncfile.density
print(density)

# To visualize the total charge wih vesta
#visu = density.visualize("vesta"); visu()

#%%
# To plot the density along the line connecting
# the first and the second in the structure:
density.plot_line(point1=0, point2=1)

#%%
# alternatively, one can define the line in terms of two points
# in fractional coordinates:
density.plot_line(point1=[0, 0, 0], point2=[2.25, 2.25, 2.25], num=300)

#%%
# To plot the density along the lines connect the firt atom in the structure
# and all the neighbors within a sphere of radius 3 Angstrom:
density.plot_line_neighbors(site_index=0, radius=3)



# coding: utf-8
"""
Interface to the GSR.nc_ file storing the Ground-state results and the electron band structure.
"""
from __future__ import annotations

import numpy as np
import pandas as pd
import pymatgen.core.units as units
import abipy.core.abinit_units as abu

from collections import OrderedDict
from typing import Optional
from tabulate import tabulate
from monty.string import list_strings, marquee
from monty.termcolor import cprint
from monty.collections import AttrDict, dict2namedtuple
from monty.functools import lazy_property
from pymatgen.core.units import ArrayWithUnit
from pymatgen.entries.computed_entries import ComputedEntry, ComputedStructureEntry
from abipy.core.mixins import AbinitNcFile, Has_Header, Has_Structure, Has_ElectronBands, NotebookWriter
from abipy.core.structure import Structure
from abipy.tools.plotting import add_fig_kwargs, get_axarray_fig_plt
from abipy.tools.typing import Figure
from abipy.abio.robots import Robot
from abipy.electrons.ebands import ElectronsReader, RobotWithEbands, ElectronBands


__all__ = [
    "GsrFile",
]

_INVALID_STRESS_TENSOR = 9999999999e+99


class GsrFile(AbinitNcFile, Has_Header, Has_Structure, Has_ElectronBands, NotebookWriter):
    """
    File containing the results of a ground-state calculation.

    Usage example:

    .. code-block:: python

        with GsrFile("foo_GSR.nc") as gsr:
            print("energy: ", gsr.energy)
            gsr.ebands.plot()

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: GsrFile
    """

    @classmethod
    def from_file(cls, filepath: str) -> GsrFile:
        """Initialize the object from a netcdf_ file."""
        return cls(filepath)

    def __init__(self, filepath: str):
        super().__init__(filepath)
        self.reader = self.r = r = GsrReader(filepath)

        # Initialize the electron bands from file
        self._ebands = r.read_ebands()

        # Add forces to structure
        if self.is_scf_run:
            self.structure.add_site_property("cartesian_forces", self.cart_forces)

    def __str__(self):
        """String representation."""
        return self.to_string()

    def to_string(self, verbose: int = 0) -> str:
        """String representation."""
        lines = []; app = lines.append

        app(marquee("File Info", mark="="))
        app(self.filestat(as_string=True))
        app("")
        app(self.structure.to_string(verbose=verbose, title="Structure"))
        if self.is_scf_run:
            app("")
            app("Stress tensor (Cartesian coordinates in GPa):\n%s" % self.cart_stress_tensor)
            #if verbose:
            #    app("Stress tensor (Cartesian coordinates in Ha/Bohr**3):\n%s" % self.cart_stress_tensor / abu.HaBohr3_GPa)
            app("")
            app("Pressure: %.3f (GPa)" % self.pressure)
            app("Energy: %.8f (eV)" % self.energy)
        app("")
        app(self.ebands.to_string(with_structure=False, verbose=verbose, title="Electronic Bands"))

        if verbose > 1:
            app("")
            app(self.hdr.to_string(verbose=verbose, title="Abinit Header"))

        return "\n".join(lines)

    @property
    def ebands(self) -> ElectronBands:
        """|ElectronBands| object."""
        return self._ebands

    @lazy_property
    def is_scf_run(self) -> bool:
        """True if the GSR has been produced by a SCF run."""
        # NOTE: We use kptopt to understand if we have a SCF/NSCF run
        # In principle one should use iscf but it's not available in the GSR.
        if "kptopt" in self.r.rootgrp.variables:
            return int(self.r.read_value("kptopt")) >= 0
        else:
            return abs(self.cart_stress_tensor[0, 0] - _INVALID_STRESS_TENSOR) > 0.1

    @lazy_property
    def ecut(self):
        """Cutoff energy in Hartree (Abinit input variable)"""
        return units.Energy(self.r.read_value("ecut"), "Ha")

    @lazy_property
    def pawecutdg(self):
        """Cutoff energy in Hartree for the PAW double grid (Abinit input variable)"""
        return units.Energy(self.r.read_value("pawecutdg"), "Ha")

    @property
    def structure(self) -> Structure:
        """|Structure| object."""
        return self.ebands.structure

    @lazy_property
    def energy(self):
        """Total energy in eV."""
        return units.Energy(self.r.read_value("etotal"), "Ha").to("eV")

    @lazy_property
    def energy_per_atom(self):
        """Total energy / number_of_atoms (eV units)"""
        return self.energy / len(self.structure)

    @lazy_property
    def cart_forces(self):
        """
        Cartesian forces in eV / Ang. None if forces are not available.
        """
        if self.is_scf_run:
            return self.r.read_cart_forces()
        return None

    @lazy_property
    def max_force(self):
        """
        Max absolute cartesian force in eV/Ang. None if forces are not available.
        """
        cart_forces = self.cart_forces
        if cart_forces is None: return None

        fmods = np.sqrt([np.dot(force, force) for force in cart_forces])
        return fmods.max()

    def force_stats(self, **kwargs):
        """
        Return a string with information on the forces.
        Return None if forces are not available.
        """
        cart_forces = self.cart_forces
        if cart_forces is None: return None

        fmods = np.sqrt([np.dot(force, force) for force in cart_forces])
        imin, imax = fmods.argmin(), fmods.argmax()

        s = "\n".join([
            "fsum: %s" % cart_forces.sum(axis=0),
            "mean: %s, std %s" % (fmods.mean(), fmods.std()),
            "minimum at site %s, cart force: %s" % (self.structure.sites[imin], cart_forces[imin]),
            "maximum at site %s, cart force: %s" % (self.structure.sites[imax], cart_forces[imax]),
        ])

        table = [["Site", "Cartesian Force", "Length"]]
        for i, fmod in enumerate(fmods):
            table.append([self.structure.sites[i], cart_forces[i], fmod])
        s += "\n" + tabulate(table)

        return s

    @lazy_property
    def cart_stress_tensor(self):
        """
        Stress tensor in GPa. Return None if not available e.g. if NSCF run.
        """
        if self.is_scf_run:
            return self.r.read_cart_stress_tensor()
        return None

    @lazy_property
    def pressure(self):
        """
        Pressure in GPa. Return None if not available e.g. if NSCF run.
        """
        if self.is_scf_run:
            pressure = - self.cart_stress_tensor.trace() / 3
            return units.FloatWithUnit(pressure, unit="GPa", unit_type="pressure")
        return None

    @lazy_property
    def residm(self):
        """
        Maximum of the residuals
        """
        return self.r.read_value("residm")

    @lazy_property
    def xc(self):
        """
        :class:`XcFunc` object with info on the exchange-correlation functional.
        Use libxc convention :cite:`Marques2012`.
        """
        return self.r.read_abinit_xcfunc()

    @lazy_property
    def energy_terms(self):
        """:class:`EnergyTerms` with the different contributions to the total energy in eV."""
        return self.r.read_energy_terms(unit="eV")

    @lazy_property
    def params(self) -> dict:
        """:class:`OrderedDict` with parameters that might be subject to convergence studies."""
        od = self.get_ebands_params()
        od["ecut"] = float(self.ecut)
        #if self.usepaw == 1
        #    od["pawecutdg"] = float(self.pawecutdg)
        return od

    def close(self) -> None:
        self.r.close()

    # FIXME: This is deprecated. Must keep it to avoid breaking ScfTask.get_results
    def as_dict(self) -> dict:
        return {}

    def get_computed_entry(self, inc_structure=True, parameters=None, data=None):
        """
        Returns a pymatgen :class:`ComputedStructureEntry` from the GSR file.
        Same API as the one used in vasp_output.get_computed_entry.

        Args:
            inc_structure (bool): Set to True if you want
                ComputedStructureEntries to be returned instead of ComputedEntries.
            parameters (list): Input parameters to include. It has to be one of
                the properties supported by the GSR object. If
                parameters is None, a default set of parameters that are
                necessary for typical post-processing will be set.
            data (list): Output data to include. Has to be one of the properties
                supported by the GSR object.

        Returns:
            ComputedStructureEntry/ComputedEntry
        """
        # TODO
        #param_names = {"is_hubbard", "hubbards", "potcar_symbols", "run_type"}
        if inc_structure:
            return ComputedStructureEntry(self.structure, self.energy,
                                          correction=0.0, parameters=parameters, data=data)
        else:
            return ComputedEntry(self.structure.composition, self.energy,
                                 parameters=parameters, data=data)

    def get_panel(self, **kwargs):
        """
        Build panel with widgets to interact with the |GsrFile| either in a notebook or in panel app.
        """
        from abipy.panels.gsr import GsrFilePanel
        return GsrFilePanel(self).get_panel(**kwargs)

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        """
        verbose = kwargs.get("verbose", 0)
        for fig in self.yield_ebands_figs(**kwargs): yield fig
        if verbose:
            for fig in self.yield_structure_figs(**kwargs): yield fig

    def yield_plotly_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of plotly figures with minimal input from the user.
        """
        verbose = kwargs.get("verbose", 0)
        for fig in self.yield_ebands_plotly_figs(**kwargs): yield fig
        if verbose:
            for fig in self.yield_structure_plotly_figs(**kwargs): yield fig

    def write_notebook(self, nbpath=None):
        """
        Write a jupyter_ notebook to ``nbpath``. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)
        first_char = "" if self.has_panel() else "#"

        nb.cells.extend([
            nbv.new_code_cell("gsr = abilab.abiopen('%s')" % self.filepath),
            nbv.new_code_cell("print(gsr)"),

            # Add panel GUI but comment the python code if panel is not available.
            nbv.new_markdown_cell("## Panel dashboard"),
            nbv.new_code_cell(f"""\
# Execute this cell to display the panel GUI (requires panel package).
# To display the dashboard inside the browser use `abiopen.py FILE --panel`.

{first_char}abilab.abipanel()
{first_char}gsr.get_panel()
"""),
            nbv.new_code_cell("gsr.ebands.plot();"),
            nbv.new_code_cell("gsr.ebands.kpoints.plot();"),
            nbv.new_code_cell("# gsr.ebands.plot_transitions(omega_ev=3.0, qpt=(0, 0, 0), atol_ev=0.1);"),
            nbv.new_code_cell("""\
if gsr.ebands.kpoints.is_ibz:
    gsr.ebands.get_edos().plot();"""),
            #nbv.new_code_cell("emass = gsr.ebands.effective_masses(spin=0, band=0, acc=4)"),
        ])

        return self._write_nb_nbpath(nb, nbpath)


class EnergyTerms(AttrDict):
    """
    Contributions to the total GS energy. See energies_type in m_energies.F90.
    """

    _NAME2DOC = OrderedDict([
        # (Name, help)
        ("e_localpsp", "Local psp energy"),
        ("e_eigenvalues", "Sum of the eigenvalues - Band energy\n" +
                          "(valid for double-counting scheme dtset%optene == 1)"),
        ("e_ewald",  "Ewald energy, store also the ion/ion energy for free boundary conditions."),
        ("e_hartree", "Hartree part of the total energy"),
        ("e_corepsp", "psp core-core energy"),
        ("e_corepspdc", "psp core-core energy double-counting"),
        ("e_kinetic", "Kinetic energy part of total energy. (valid for direct scheme, dtset%optene == 0"),
        ("e_nonlocalpsp", "Nonlocal pseudopotential part of total energy."),
        ("e_entropy", "Entropy energy due to the occupation number smearing (if metal)\n" +
                      "Value is multiplied by dtset%tsmear, see %entropy for the entropy alone\n." +
                      "(valid for metals, dtset%occopt>=3 .and. dtset%occopt<=8)"),
        ("entropy", "Entropy term"),
        ("e_xc", "Exchange-correlation energy"),
        #("e_vdw_dftd2", "Dispersion energy from DFT-D2 Van der Waals correction"),
        ("e_xcdc", "enxcdc=exchange-correlation double-counting energy"),
        ("e_paw", "PAW spherical part energy"),
        ("e_pawdc", "PAW spherical part double-counting energy"),
        ("e_elecfield", "Electric enthalpy, by adding both ionic and electronic contributions"),
        ("e_magfield", "Orbital magnetic enthalpy, by adding orbital contribution"),
        ("e_fermie", "Fermie energy"),
        ("e_sicdc", "Self-interaction energy double-counting"),
        ("e_exactX", "Fock exact-exchange energy"),
        ("h0", "h0=e_kinetic+e_localpsp+e_nonlocalpsp"),
        ("e_electronpositron", "Electron-positron: electron-positron interaction energy"),
        ("edc_electronpositron", "Electron-positron: double-counting electron-positron interaction energy"),
        ("e0_electronpositron", "Electron-positron: energy only due to unchanged particles\n" +
                                "(if calctype=1, energy due to electrons only)\n" +
                                "(if calctype=2, energy due to positron only)\n"),
        ("e_monopole", "Monopole correction to the total energy for charged supercells"),
        # FIXME: Some names have been changed in Abinit8, I should recheck the code.
        #("e_xc_vdw", "vdW-DF correction to the XC energy"),
    ])

    ALL_KEYS = list(_NAME2DOC.keys())

    def __str__(self):
        return self.to_string(with_doc=False)

    __repr__ = __str__

    def to_string(self, verbose: int = 0, with_doc: bool = True) -> str:
        """String representation, with documentation if with_doc."""
        lines = [str(self.table)]
        if with_doc:
            for k, doc in self._NAME2DOC.items():
                lines.append("%s: %s" % (k, doc))

        return "\n".join(lines)

    @property
    def table(self):
        """string with results in tabular form."""
        table = [["Term", "Value"]]
        for k, doc in self._NAME2DOC.items():
            table.append([k, self[k]])
        return tabulate(table, tablefmt="plain")

    #def get_dataframe(self):
    #    """Return a |pandas-DataFrame|"""
    #    d = {k: float(self[k]) for k in self}
    #    return pd.DataFrame(d, index=[None], columns=list(d.keys()))
    #    #return pd.DataFrame(d, columns=list(d.keys()))


class GsrReader(ElectronsReader):
    """
    This object reads the results stored in the _GSR (Ground-State Results) file produced by ABINIT.
    It provides helper function to access the most important quantities.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: GsrReader
    """
    def read_cart_forces(self, unit="eV ang^-1"):
        """
        Read and return a |numpy-array| with the cartesian forces in unit ``unit``.
        Shape (natom, 3)
        """
        return ArrayWithUnit(self.read_value("cartesian_forces"), "Ha bohr^-1").to(unit)

    def read_cart_stress_tensor(self):
        """
        Return the stress tensor (3x3 matrix) in cartesian coordinates in GPa.
        If MaskedArray (i.e. tensor was not computed  e.g. Nscf run) set it to _INVALID_STRESS_TENSOR
        """
        # Abinit stores 6 unique components of this symmetric 3x3 tensor:
        # Given in order (1,1), (2,2), (3,3), (3,2), (3,1), (2,1).
        c = self.read_value("cartesian_stress_tensor")
        tensor = np.empty((3, 3), dtype=float)

        if np.ma.is_masked(c[()]):
            # NSCF
            tensor.fill(_INVALID_STRESS_TENSOR)
        else:
            for i in range(3): tensor[i, i] = c[i]
            for p, (i, j) in enumerate(((2, 1), (2, 0), (1, 0))):
                tensor[i, j] = c[3 + p]
                tensor[j, i] = c[3 + p]
            tensor *= abu.HaBohr3_GPa

        from abipy.tools.tensors import Stress
        return Stress(tensor)

    def read_energy_terms(self, unit="eV"):
        """
        Return a dictionary with the different contributions to the total electronic energy.
        """
        convert = lambda e: units.Energy(e, unit="Ha").to(unit)
        d = OrderedDict()
        for k in EnergyTerms.ALL_KEYS:
            if k == "e_nonlocalpsp" and k not in self.rootgrp.variables:
                # Renamed in 8.9
                d[k] = convert(self.read_value("e_nlpsp_vfock"))
            else:
                d[k] = convert(self.read_value(k))

        return EnergyTerms(**d)


class GsrRobot(Robot, RobotWithEbands):
    """
    This robot analyzes the results contained in multiple GSR.nc_ files.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: GsrRobot
    """
    EXT = "GSR"

    def get_dataframe(self, with_geo=True, abspath=False, with_paths=True, funcs=None, **kwargs) -> pd.DataFrame:
        """
        Return a |pandas-DataFrame| with the most important GS results and the filenames as index.

        Args:
            with_geo: True if structure info should be added to the dataframe
            abspath: True if paths in the index should be absolute. Default: Relative to getcwd().
            with_paths: False if filepaths should not be added

        kwargs:
            attrs:
                List of additional attributes of the |GsrFile| to add to the DataFrame.
            funcs: Function or list of functions to execute to add more data to the DataFrame.
                Each function receives a |GsrFile| object and returns a tuple (key, value)
                where key is a string with the name of column and value is the value to be inserted.
        """
        # Add attributes specified by the users
        # TODO add more columns
        attrs = [
            "energy", "energy_per_atom", "pressure", "max_force",
            "ecut", "pawecutdg", "tsmear", "nkpt",
            "nsppol", "nspinor", "nspden",
        ] + kwargs.pop("attrs", [])

        rows, row_names = [], []
        for label, gsr in self.items():
            row_names.append(label)
            d = OrderedDict()

            # Add info on structure.
            if with_geo:
                d.update(gsr.structure.get_dict4pandas(with_spglib=True))

            for aname in attrs:
                if aname == "nkpt":
                    value = len(gsr.ebands.kpoints)
                else:
                    value = getattr(gsr, aname, None)
                    if value is None: value = getattr(gsr.ebands, aname, None)
                d[aname] = value

            # Execute functions
            if funcs is not None: d.update(self._exec_funcs(funcs, gsr))
            rows.append(d)

        index = None
        if with_paths:
            index = row_names if not abspath else self._to_relpaths(row_names)
        return pd.DataFrame(rows, index=index, columns=list(rows[0].keys()))

    def get_eos_fits_dataframe(self, eos_names="murnaghan"):
        """
        Fit energy as function of volume to get the equation of state,
        equilibrium volume, bulk modulus and its derivative wrt to pressure.

        Args:
            eos_names: String or list of strings with EOS names.
                For the list of available models, see pymatgen.analysis.eos.

        Return:
            (fits, dataframe) namedtuple.
                fits is a list of ``EOSFit object``
                dataframe is a |pandas-DataFrame| with the final results.
        """
        # Read volumes and energies from the GSR files.
        energies, volumes = [], []
        for label, gsr in self.items():
            energies.append(float(gsr.energy))
            volumes.append(float(gsr.structure.volume))

        # Order data by volumes if needed.
        if np.any(np.diff(volumes) < 0):
            ves = sorted(zip(volumes, energies), key=lambda t: t[0])
            volumes = [t[0] for t in ves]
            energies = [t[1] for t in ves]

        # Note that eos.fit expects lengths in Angstrom, and energies in eV.
        # I'm also monkey-patching the plot method.
        from pymatgen.analysis.eos import EOS
        if eos_names == "all":
            # Use all the available models.
            eos_names = [n for n in EOS.MODELS if n not in ("deltafactor", "numerical_eos")]
        else:
            eos_names = list_strings(eos_names)

        fits, index, rows = [], [], []
        for eos_name in eos_names:
            try:
                fit = EOS(eos_name=eos_name).fit(volumes, energies)
            except Exception as exc:
                cprint("EOS %s raised exception:\n%s" % (eos_name, str(exc)))
                continue

            # Replace plot with plot_ax method
            fit.plot = fit.plot_ax
            fits.append(fit)
            index.append(eos_name)
            rows.append(OrderedDict([(aname, getattr(fit, aname)) for aname in
                ("v0", "e0", "b0_GPa", "b1")]))

        dataframe = pd.DataFrame(rows, index=index, columns=list(rows[0].keys()) if rows else None)
        return dict2namedtuple(fits=fits, dataframe=dataframe)

    def get_energyterms_dataframe(self, iref: Optional[int] = None) -> pd.DataFrame:
        """
        Build and return dataframe with the different contributions to the total energy in eV

        Args:
            iref: Index of the abifile used as reference: the energies of the
            ``iref`` gsrfile will be subtracted from the other rows. Ignored if None.

        Return: |pandas-Dataframe|
        """
        rows, row_names = [], []
        for label, gsr in self.items():
            row_names.append(label)
            # Add total energy.
            d = OrderedDict([("energy", gsr.energy)])
            d.update(gsr.energy_terms)
            d.update(gsr.params)
            rows.append(d)

        df = pd.DataFrame(rows, index=row_names, columns=list(rows[0].keys()))
        if iref is not None:
            # Subtract iref row from the rest of the rows.
            iref_row = df.iloc[[iref]].values[0]
            df = df.apply(lambda row: row - iref_row, axis=1)

        return df

    @add_fig_kwargs
    def gridplot_eos(self, eos_names="all", fontsize=6, **kwargs) -> Figure:
        """
        Plot multiple EOS on a grid with captions showing the final results.

        Args:
            eos_names: String or list of strings with EOS names. See pymatgen.analysis.EOS
            fontsize: Fontsize used for caption text.

        Returns: |matplotlib-Figure|
        """
        r = self.get_eos_fits_dataframe(eos_names=eos_names)

        num_plots, ncols, nrows = len(r.fits), 1, 1
        if num_plots > 1:
            ncols = 2
            nrows = (num_plots // ncols) + (num_plots % ncols)

        # Build grid of plots.
        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                                sharex=False, sharey=False, squeeze=False)
        ax_list = ax_list.ravel()

        for i, (fit, ax) in enumerate(zip(r.fits, ax_list)):
            fit.plot_ax(ax=ax, fontsize=fontsize, label="", show=False)

        # Get around a bug in matplotlib
        if num_plots % ncols != 0:
            ax_list[-1].axis('off')

        return fig

    @add_fig_kwargs
    def plot_gsr_convergence(self, sortby=None, hue=None, fontsize=8,
                             items=("energy", "pressure", "max_force"), **kwargs) -> Figure:
        """
        Plot the convergence of the most important quantities available in the GSR file
        wrt to the ``sortby`` parameter. Values can be optionally grouped by ``hue``.

        Args:
            sortby: Define the convergence parameter, sort files and produce plot labels.
                Can be None, string or function. If None, no sorting is performed.
                If string and not empty it's assumed that the abifile has an attribute
                with the same name and `getattr` is invoked.
                If callable, the output of `sortby(abifile)` is used.
            hue: Variable that define subsets of the data, which will be drawn on separate lines.
                Accepts callable or string
                If string, it's assumed that the abifile has an attribute with the same name and getattr is invoked.
                If callable, the output of `hue(abifile)` is used.
            items: List of GSR attributes (or callables) to be analyzed.
            fontsize: legend and label fontsize.

        Returns: |matplotlib-Figure|

        Example::

             gsr.plot_gsr_convergence(sortby="ecut")
             gsr.plot_gsr_convergence(sortby="nkpt", hue="tsmear")
        """
        return self.plot_convergence_items(items, sortby=sortby, hue=hue,
                                           fontsize=fontsize, show=False, **kwargs)

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        Used in abiview.py to get a quick look at the results.
        """
        yield self.plot_lattice_convergence(show=False)
        yield self.plot_gsr_convergence(show=False)
        for fig in self.get_ebands_plotter().yield_figs(): yield fig

    def get_panel(self, **kwargs):
        """
        Build panel with widgets to interact with the |GsrRobot| either in a notebook or in panel app.
        """
        from abipy.panels.gsr import GsrRobotPanel
        return GsrRobotPanel(robot=self).get_panel(**kwargs)

    def write_notebook(self, nbpath=None) -> str:
        """
        Write a jupyter_ notebook to ``nbpath``. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        args = [(l, f.filepath) for l, f in self.items()]
        nb.cells.extend([
            #nbv.new_markdown_cell("# This is a markdown cell"),
            nbv.new_code_cell("robot = abilab.GsrRobot(*%s)\nrobot.trim_paths()\nrobot" % str(args)),
            nbv.new_code_cell("ebands_plotter = robot.get_ebands_plotter()"),
            nbv.new_code_cell("df = ebands_plotter.get_ebands_frame()\ndisplay(df)"),
            nbv.new_code_cell("ebands_plotter.ipw_select_plot()"),
            nbv.new_code_cell("#anim = ebands_plotter.animate();"),
            nbv.new_code_cell("edos_plotter = robot.get_edos_plotter()"),
            nbv.new_code_cell("edos_plotter.ipw_select_plot()"),
            nbv.new_code_cell("#robot.gridplot_eos();"),
        ])

        # Mixins
        nb.cells.extend(self.get_baserobot_code_cells())
        nb.cells.extend(self.get_ebands_code_cells())

        return self._write_nb_nbpath(nb, nbpath)



#!/usr/bin/env python
"""Analyze the parallel efficiency of the RPA code (hilbert transform, 60 frequencies and gwpara==2)"""
import sys
import abipy.abilab as abilab
import abipy.data as abidata
import abipy.flowtk as flowtk

from itertools import product
from abipy.benchmarks import bench_main, BenchmarkFlow


def make_inputs(paw=False):
    # Crystalline silicon
    # Calculation of the GW correction to the direct band gap in Gamma
    # Dataset 1: ground state calculation
    # Dataset 2: NSCF calculation
    # Dataset 3: calculation of the screening
    # Dataset 4: calculation of the Self-Energy matrix elements (GW corrections)
    structure = abilab.Structure.from_file(abidata.cif_file("si.cif"))
    pseudos = abidata.pseudos("14si.pspnc") if not paw else abidata.pseudos("Si.GGA_PBE-JTH-paw.xml")

    multi = abilab.MultiDataset(structure, pseudos=pseudos, ndtset=4)

    ecut = 24
    multi.set_vars(
        ecut=ecut,
        pawecutdg=ecut*4 if paw else None,
        timopt=-1,
        istwfk="*1",
        paral_kgb=0,
    )

    gs, nscf, scr, sigma = multi.split_datasets()

    # This grid is the most economical, but does not contain the Gamma point.
    ngkpt = [5, 5, 5]
    gs_kmesh = dict(
        ngkpt=ngkpt,
        shiftk=[0.5, 0.5, 0.5,
                0.5, 0.0, 0.0,
                0.0, 0.5, 0.0,
                0.0, 0.0, 0.5]
    )

    # This grid contains the Gamma point, which is the point at which
    # we will compute the (direct) band gap.
    gw_kmesh = dict(
        ngkpt=ngkpt,
        shiftk=[0.0, 0.0, 0.0,
                0.0, 0.5, 0.5,
                0.5, 0.0, 0.5,
                0.5, 0.5, 0.0]
    )

    # Dataset 1 (GS run)
    gs.set_kmesh(**gs_kmesh)
    gs.set_vars(tolvrs=1e-6,
                nband=8,
                )

    # Dataset 2 (NSCF run)
    # Here we select the second dataset directly with the syntax inp[2]
    nscf.set_kmesh(**gw_kmesh)
    nscf.set_vars(iscf=-2,
                  tolwfr=1e-8,
                  nband=600,
                  nbdbuf=200,
                  )

    # Dataset3: Calculation of the screening.
    scr.set_kmesh(**gw_kmesh)
    scr.set_vars(
        optdriver=3,
        gwcalctyp=9,
        gwpara=2,
        nband=25,
        ecutwfn=ecut,
        symchi=1,
        awtr=2,
        inclvkb=1,
        ecuteps=6.0,
        spmeth=1,        # Use Hilbert transform : Im chi0 --> chi0.
        nomegasf=250,    # Number of points for Imchi0
        nfreqre=50,
        nfreqim=10,
        freqremax="25 eV",
        freqremin="0 eV",
    )

    return gs, nscf, scr


def build_flow(options):
    """
    Build an `AbinitWorkflow` used for benchmarking ABINIT.
    """
    gs_inp, nscf_inp, scr_inp = make_inputs(paw=options.paw)
    flow = BenchmarkFlow(workdir=options.get_workdir(__file__), remove=options.remove)

    bands = flowtk.BandStructureWork(gs_inp, nscf_inp)
    flow.register_work(bands)
    flow.exclude_from_benchmark(bands)

    #for nband in [200, 400, 600]:
    for nband in [600]:
        scr_work = flowtk.Work()
        inp = scr_inp.new_with_vars(nband=nband)
        mpi_list = options.mpi_list
        if mpi_list is None:
            # Cannot call autoparal here because we need a WFK file.
            print("Using hard coded values for mpi_list")
            mpi_list = [np for np in range(1, nband+1) if abs((nband - 4) % np) < 1]
        if options.verbose: print("Using nband %d and mpi_list: %s" % (nband, mpi_list))

        for mpi_procs, omp_threads in product(mpi_list, options.omp_list):
            if not options.accept_mpi_omp(mpi_procs, omp_threads): continue
            manager = options.manager.new_with_fixed_mpi_omp(mpi_procs, omp_threads)
            scr_work.register_scr_task(inp, manager=manager, deps={bands.nscf_task: "WFK"})

        flow.register_work(scr_work)

    return flow.allocate()


@bench_main
def main(options):
    if options.info:
        # print doc string and exit.
        print(__doc__)
        return
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



"""Panels to interact with the AbiPy tasks."""
from __future__ import annotations

import panel as pn
#import panel.widgets as pnw

from abipy.panels.viewers import AceViewer
#from abipy.panels.core import mpl, ply, dfc, depends_on_btn_click, Loading
from abipy.panels.nodes import NodeParameterized
from abipy.flowtk.tasks import AbinitTask


class TaskPanel(NodeParameterized):
    """
    Provides widgets to interact with an AbiPy Task.
    """

    def __init__(self, task: AbinitTask, **params):
        NodeParameterized.__init__(self, node=task, **params)
        self.task = task

        #self.structures_btn = pnw.Button(name="Show Structures", button_type='primary')
        #self.structures_io_checkbox = pnw.CheckBoxGroup(
        #    name='Input/Output Structure', value=['output'], options=['input', 'output'], inline=Tru

    def get_inputs_view(self):
        """
        Show the input files of the task: input file, submission script and TaskManager
        """
        file = self.task.job_file
        text = file.read() if file.exists else "Cannot find job_file!"
        job_file = pn.pane.Markdown(f"```shell\n{text}\n```")

        from .viewers import JSONViewer
        json_view = JSONViewer(self.task.manager.as_dict())

        def card(title, *items, collapsed=True):
            return pn.Card(*items,
                           title=title,
                           collapsed=collapsed,
                           sizing_mode="stretch_width",
                           header_color="blue",
                           #header_background="blue",
                           )

        return pn.Column(
            f"## Input files of `{repr(self.task)}`",
            card("Input file", self.html_with_clipboard_btn(self.task.input), collapsed=False),
            pn.layout.Divider(),
            card("Submission script", job_file),
            pn.layout.Divider(),
            card("TaskManager", json_view),
            pn.layout.Divider(),
            sizing_mode="stretch_width",
        )

    def get_errs_view(self):
        """
        Show the error files of the task
        Return None if no error is found so that we don't show this view in the GUI.
        """
        col = pn.Column(sizing_mode="stretch_width"); cext = col.extend

        count = 0
        for fname in ("stderr_file", "mpiabort_file", "qerr_file", "qout_file"):
            file = getattr(self.task, fname)
            if file.exists:
                text = file.read().strip()
                if text:
                    cext([f"## {fname}",
                         pn.pane.Markdown(f"```shell\n{text}\n```"),
                         pn.layout.Divider()
                         ])
                    count += 1

        return col if count > 0 else None

    def get_main_text_outs_view(self) -> pn.Column:
        """
        Show the main text output files of the task.
        """
        col = pn.Column(f"## Main output and log file of `{repr(self.task)}`",
                        sizing_mode="stretch_width")

        for fname in ("output_file", "log_file"):
            file = getattr(self.task, fname)
            col.append(AceViewer(file.path))

        return col

    #@depends_on_btn_click("structures_btn")
    #def on_structures_btn(self):
    #    what = ""
    #    if "input" in self.structures_io_checkbox.value: what += "i"
    #    if "output" in self.structures_io_checkbox.value: what += "o"
    #    dfs = self.flow.compare_structures(nids=self.nids,
    #                                       what=what,
    #                                       verbose=self.verbose, with_spglib=False, printout=False,
    #                                       with_colors=False)

    #    return pn.Row(dfc(dfs.lattice), sizing_mode="scale_width")

    def get_panel(self, as_dict=False, **kwargs):
        """Return tabs with widgets to interact with the flow."""
        d = dict()

        # This stuff is computed lazyly when the tab is activated.
        d["Inputs"] = self.get_inputs_view()
        d["Output"] = self.get_main_text_outs_view()
        view = self.get_errs_view()
        if view is not None: d["ErrFiles"] = view

        super_d =  super().get_panel(as_dict=True)
        d.update(super_d)
        ##d["Structures"] = pn.Row(pn.Column(self.structures_io_checkbox, self.structures_btn), self

        if as_dict: return d

        return self.get_template_from_tabs(d, template=kwargs.get("template", None))


#!/usr/bin/env python
"""
Script to analyze performance
"""
from __future__ import annotations

import sys
import os
import argparse
import subprocess
import abipy.tools.cli_parsers as cli

from pprint import pprint
from shutil import which
from monty.termcolor import cprint
from monty.functools import prof_main
from abipy import abilab


def get_epilog() -> str:
    s = """\
======================================================================================================
Usage example:

    abiopen.py FILE          => Open file in ipython shell.

======================================================================================================

"""
    return s


def get_parser(with_epilog=False):
    parser = argparse.ArgumentParser(epilog=get_epilog() if with_epilog else "",
                                     formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('--loglevel', default="ERROR", type=str,
        help="Set the loglevel. Possible values: CRITICAL, ERROR (default), WARNING, INFO, DEBUG")
    parser.add_argument('-V', '--version', action='version', version=abilab.__version__)

    parser.add_argument('-v', '--verbose', default=0, action='count', # -vv --> verbose=2
        help='verbose, can be supplied multiple times to increase verbosity')

    parser.add_argument("filepath", help="File to open. See table below for the list of supported extensions.")

    # notebook options.
    #parser.add_argument('-nb', '--notebook', action='store_true', default=False, help="Open file in jupyter notebook")
    #parser.add_argument('--classic-notebook', "-cnb", action='store_true', default=False,
    #                    help="Use classic jupyter notebook instead of jupyterlab.")
    #parser.add_argument('--no-browser', action='store_true', default=False,
    #                    help=("Start the jupyter server to serve the notebook "
    #                          "but don't open the notebook in the browser.\n"
    #                          "Use this option to connect remotely from localhost to the machine running the kernel"))
    #parser.add_argument('--foreground', action='store_true', default=False,
    #                    help="Run jupyter notebook in the foreground.")

    # Expose option.
    #parser.add_argument('-e', '--expose', action='store_true', default=False,
    #    help="Open file and generate matplotlib figures automatically by calling expose method.")
    #parser.add_argument("-s", "--slide-mode", default=False, action="store_true",
    #    help="Iterate over figures. Expose all figures at once if not given on the CLI.")
    #parser.add_argument("-t", "--slide-timeout", type=int, default=None,
    #    help="Close figure after slide-timeout seconds (only if slide-mode). Block if not specified.")
    #parser.add_argument('-sns', "--seaborn", const="paper", default=None, action='store', nargs='?', type=str,
    #    help='Use seaborn settings. Accept value defining context in ("paper", "notebook", "talk", "poster").')
    #parser.add_argument('-mpl', "--mpl-backend", default=None,
    #    help=("Set matplotlib interactive backend. "
    #          "Possible values: GTKAgg, GTK3Agg, GTK, GTKCairo, GTK3Cairo, WXAgg, WX, TkAgg, Qt4Agg, Qt5Agg, macosx."
    #          "See also: https://matplotlib.org/faq/usage_faq.html#what-is-a-backend."))
    #parser.add_argument("-ew", "--expose-web", default=False, action="store_true",
    #        help='Generate matplotlib plots in $BROWSER instead of X-server. WARNING: Not all the features are supported.')
    #parser.add_argument("-ply", "--plotly", default=False, action="store_true",
    #        help='Generate plotly plots in $BROWSER instead of matplotlib. WARNING: Not all the features are supported.')

    return parser


@prof_main
def main():
    def show_examples_and_exit(err_msg=None, error_code=1):
        """Display the usage of the script."""
        sys.stderr.write(get_epilog())
        if err_msg:
            sys.stderr.write("Fatal Error\n" + err_msg + "\n")
        sys.exit(error_code)

    parser = get_parser(with_epilog=True)

    # Parse the command line.
    try:
        options = parser.parse_args()
    except Exception:
        show_examples_and_exit(error_code=1)

    cli.set_loglevel(options.loglevel)

    ##############################################################################################
    # Handle meta options i.e. options that set other options.
    # OK, it's not very clean but I haven't find any parse API to express this kind of dependency.
    ##############################################################################################
    if options.plotly: options.expose = True
    if options.expose_web: options.expose = True

    if options.verbose > 2: print(options)

    # Set matplotlib backend
    if options.mpl_backend is not None:
        import matplotlib
        matplotlib.use(options.mpl_backend)

    # Use seaborn settings.
    if options.seaborn:
        import seaborn as sns
        sns.set(context=options.seaborn, style='darkgrid', palette='deep',
                font='sans-serif', font_scale=1, color_codes=False, rc=None)

    if not os.path.exists(options.filepath):
        raise RuntimeError("%s: no such file" % options.filepath)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
"""
Gruneisen with Phonopy and AbiPy
================================

Compute Gruneisedn parameters with phonopy (supercells and finite-difference method).
"""
from __future__ import print_function, division, unicode_literals, absolute_import

import sys
import os
import abipy.abilab as abilab
import abipy.data as abidata
import abipy.flowtk as flowtk

from abipy.flowtk.abiphonopy import PhonopyGruneisenWork


def build_flow(options):
    """
    Create a `Flow` for phonon calculations with phonopy:
    """
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        __file__ = os.path.join(os.getcwd(), "run_gruneisen_phonopy_si.py")
        options.workdir = os.path.basename(__file__).replace(".py", "").replace("run_", "flow_")

    # Initialize structure and pseudos
    structure = abilab.Structure.from_file(abidata.cif_file("si.cif"))
    pseudos = abidata.pseudos("14si.pspnc")

    # Build input for GS calculation.
    gsinp = abilab.AbinitInput(structure, pseudos)
    gsinp.set_vars(ecut=8, nband=4, toldff=1.e-6)

    # This gives ngkpt = 4x4x4 with 4 shifts for the initial unit cell.
    # The k-point sampling will be rescaled when we build the supercell in PhonopyWork.
    gsinp.set_autokmesh(nksmall=4)

    flow = flowtk.Flow(workdir=options.workdir)

    # Use a 2x2x2 supercell to compute phonons with phonopy
    work = PhonopyGruneisenWork.from_gs_input(gsinp, voldelta=0.01, scdims=[2, 2, 2])
    flow.register_work(work)

    return flow


# This block generates the thumbnails in the Abipy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
r"""
Optic Flow
==========

Optical spectra with Optic.
"""
from __future__ import print_function, division, unicode_literals, absolute_import

import sys
import os
import abipy.data as abidata
import abipy.abilab as abilab
import abipy.flowtk as flowtk


def build_flow(options, paral_kgb=0):
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(__file__).replace(".py", "").replace("run_", "flow_")

    multi = abilab.MultiDataset(structure=abidata.structure_from_ucell("GaAs"),
                                pseudos=abidata.pseudos("31ga.pspnc", "33as.pspnc"), ndtset=5)

    # Global variables
    kmesh = dict(ngkpt=[4, 4, 4],
                 nshiftk=4,
                 shiftk=[[0.5, 0.5, 0.5],
                         [0.5, 0.0, 0.0],
                         [0.0, 0.5, 0.0],
                         [0.0, 0.0, 0.5]]
                )

    global_vars = dict(ecut=2, paral_kgb=paral_kgb)
    global_vars.update(kmesh)

    multi.set_vars(global_vars)

    # Dataset 1 (GS run)
    multi[0].set_vars(
        tolvrs=1e-6,
        nband=4,
    )

    # NSCF run with large number of bands, and points in the the full BZ
    multi[1].set_vars(
        iscf=-2,
        nband=20,
        nstep=25,
        kptopt=1,
        tolwfr=1.e-9,
        #kptopt=3,
    )

    # Fourth dataset : ddk response function along axis 1
    # Fifth dataset : ddk response function along axis 2
    # Sixth dataset : ddk response function along axis 3
    for idir in range(3):
        rfdir = 3 * [0]
        rfdir[idir] = 1

        multi[2+idir].set_vars(
            iscf=-3,
            nband=20,
            nstep=1,
            nline=0,
            prtwf=3,
            kptopt=3,
            nqpt=1,
            qpt=[0.0, 0.0, 0.0],
            rfdir=rfdir,
            rfelfd=2,
            tolwfr=1.e-9,
        )

    scf_inp, nscf_inp, ddk1, ddk2, ddk3 = multi.split_datasets()

    # Initialize the flow.
    flow = flowtk.Flow(options.workdir, manager=options.manager)

    bands_work = flowtk.BandStructureWork(scf_inp, nscf_inp)
    flow.register_work(bands_work)

    ddk_work = flowtk.Work()
    for inp in [ddk1, ddk2, ddk3]:
        ddk_work.register_ddk_task(inp, deps={bands_work.nscf_task: "WFK"})

    flow.register_work(ddk_work)

    # Optic does not support MPI with ncpus > 1.
    optic_input = abilab.OpticInput(
        broadening=0.002,          # Value of the smearing factor, in Hartree
        domega=0.0003,             # Frequency mesh.
        maxomega=0.3,
        scissor=0.000,             # Scissor shift if needed, in Hartree
        tolerance=0.002,           # Tolerance on closeness of singularities (in Hartree)
        num_lin_comp=1,            # Number of components of linear optic tensor to be computed
        lin_comp=11,               # Linear coefficients to be computed (x=1, y=2, z=3)
        num_nonlin_comp=2,         # Number of components of nonlinear optic tensor to be computed
        nonlin_comp=(123, 222),    # Non-linear coefficients to be computed
    )

    # TODO
    # Check is the order of the 1WF files is relevant. Can we use DDK files ordered
    # in an arbitrary way or do we have to pass (x,y,z)?
    optic_task = flowtk.OpticTask(optic_input, nscf_node=bands_work.nscf_task, ddk_nodes=ddk_work)
    flow.register_task(optic_task)

    return flow


def optic_flow_from_files():
    # Optic does not support MPI with ncpus > 1.
    manager = flowtk.TaskManager.from_user_config()
    manager.set_mpi_procs(1)

    flow = flowtk.Flow(workdir="OPTIC_FROM_FILE", manager=manager)

    ddk_nodes = [
        "/Users/gmatteo/Coding/abipy/abipy/data/runs/OPTIC/work_1/task_0/outdata/out_1WF",
        "/Users/gmatteo/Coding/abipy/abipy/data/runs/OPTIC/work_1/task_1/outdata/out_1WF",
        "/Users/gmatteo/Coding/abipy/abipy/data/runs/OPTIC/work_1/task_2/outdata/out_1WF",
    ]
    nscf_node = "/Users/gmatteo/Coding/abipy/abipy/data/runs/OPTIC/work_0/task_1/outdata/out_WFK"

    optic_task = flowtk.OpticTask(optic_input, nscf_node=nscf_node, ddk_nodes=ddk_nodes)
    flow.register_task(optic_task)
    return flow


# This block generates the thumbnails in the Abipy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).plot_networkx(tight_layout=True)


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



""""Basic tools and mixin classes for AbiPy panels."""
from __future__ import annotations

import io
import sys
import tempfile
import functools
import textwrap
import time
import traceback
import shutil
import param
import numpy as np
import pandas as pd
import bokeh.models.widgets as bkw
import panel as pn
import panel.widgets as pnw

from monty.functools import lazy_property
from monty.termcolor import cprint
from abipy.core import abinit_units as abu
from abipy.core.structure import Structure
from abipy.tools.plotting import push_to_chart_studio
from abipy.tools.decorators import Appender


_ABINIT_TEMPLATE_NAME = "FastList"


def set_abinit_template(template_name: str):
    global _ABINIT_TEMPLATE_NAME
    _ABINIT_TEMPLATE_NAME = template_name


def get_abinit_template_cls_kwds() -> tuple:
    cls =  get_template_cls_from_name(_ABINIT_TEMPLATE_NAME)
    kwds = dict(header_background="#ff8c00", # Dark orange
                favicon="/assets/img/abinit_favicon.ico",
                logo="/assets/img/abipy_logo.png", # TODO: Need new panel version to fix logo alignment in FastLIst.
                #sidebar_footer (str): Can be used to insert additional HTML.
                #                      For example a menu, some additional info, links etc.
                #enable_theme_toggle=False,  # If True a switch to toggle the Theme is shown. Default is True.
                )

    return cls, kwds


def open_html(html_string: str, browser: str = None) -> None:
    """
    Open a string with an HTML document in browser.
    """
    import tempfile
    import webbrowser
    with tempfile.NamedTemporaryFile(mode="wt", suffix=".html", delete=False) as tmpfile:
        tmpfile.write(html_string)
        webbrowser.get(browser).open_new_tab(f"file://{tmpfile.name}")


def abipanel(panel_template: str = "FastList"):
    """
    Activate panel extensions used by AbiPy. Return panel module.

    Args:
        panel_template: String with the name of the panel template to be used by default.
    """
    try:
        import panel as pn
    except ImportError as exc:
        cprint("Use `conda install panel` or `pip install panel` to install the python package.", "red")
        raise exc

    set_abinit_template(panel_template)

    #pn.extension(loading_spinner='dots', loading_color='#00aa41')

    pn.extension(notifications=True)
    #pn.config.notifications = True
    #pn.state.notifications.position = 'top-right'

    extensions = [
        "plotly",
        #"katex",
        "mathjax",
        "terminal",
        "tabulator",
        "ace",   # NB: This enters in conflict with Abipy Book
        #"gridstack",
        #"ipywidgets",
    ]

    css_files = [
        # FIXME
        #pn.io.resources.CSS_URLS['font-awesome'],
    ]

    #pn.extension(loading_spinner='petal', loading_color='#00aa41')
    #print("loading extensions:", extensions)

    #import os
    #abipy_css = os.path.join(os.path.dirname(__file__), "assets", "css", "abipy.css")

    pn.extension(*extensions, css_files=css_files) #, css_files=[abipy_css])
    #pn.extension(template='fast', theme='dark')

    pn.config.js_files.update({
        # This for copy to clipboard.
        "clipboard": "https://cdn.jsdelivr.net/npm/clipboard@2/dist/clipboard.min.js",
        # This for the jsmol viewer.
        "jsmol": "https://chemapps.stolaf.edu/jmol/jsmol/JSmol.min.js",
    })

    #pn.extension('ipywidgets')

    #pn.config.js_files.update({
    #    'ngl': 'https://cdn.jsdelivr.net/gh/arose/ngl@v2.0.0-dev.33/dist/ngl.js',
    #})
    #pn.extension(comms='ipywidgets')
    #pn.config.sizing_mode = "stretch_width"

    #pn.config.css_files.extend([
    #    "https://maxcdn.bootstrapcdn.com/font-awesome/4.5.0/css/font-awesome.min.css",
    #])

    #css = """
    #    .pnx-file-upload-area input[type=file] {
    #        width: 100%;
    #        height: 100%;
    #        border: 3px dashed #9E9E9E;
    #        background: transparent;
    #        border-radius: 5px;
    #        text-align: left;
    #        margin: auto;
    #    }
    #"""

    #css = open(abipy_css, "rt").read()
    #pn.config.raw_css.append(css)

    return pn


def gen_id(n: int = 1, pre: str = "uuid-"):
    """
    Generate ``n`` universally unique identifiers prepended with ``pre`` string.
    Return string if n == 1 or list of strings if n > 1
    """
    # The HTML4 spec says:
    # ID and NAME tokens must begin with a letter ([A-Za-z]) and may be followed by any number of letters,
    # digits ([0-9]), hyphens ("-"), underscores ("_"), colons (":"), and periods (".").
    import uuid
    if n == 1:
        return pre + str(uuid.uuid4())
    elif n > 1:
        return [pre + str(uuid.uuid4()) for i in range(n)]
    else:
        raise ValueError("n must be > 0 but got %s" % str(n))


def get_template_cls_from_name(name: str):
    """
    Return panel template from string.
    Support name in the form `FastList` as well as `FastListTemplate`.
    """
    # Example: pn.template.FastGridTemplate or pn.template.GoldenTemplate
    if hasattr(pn.template, name):
        return getattr(pn.template, name)

    try_name = name + "Template"
    if hasattr(pn.template, try_name):
        return getattr(pn.template, try_name)

    raise ValueError(f"""
Don't know how to return panel template from string: `{name}`.
Possible templates are: {list(pn.template.__dict__.keys())}
""")


add_mp_rest_docstring = Appender("""
NB: Requires a String API key for accessing the MaterialsProject REST interface.
Please apply on the Materials Project website for one.
If this is None, the code will check if there is a `PMG_MAPI_KEY` in your .pmgrc.yaml.
If so, it will use that environment
This makes easier for heavy users to simply add this environment variable to their setups and MPRester can
then be called without any arguments.
""", indents=0)


def depends_on_btn_click(btn_name: str,
                         show_doc: bool = True,
                         show_shared_wdg_warning: bool = True,
                         show_exc: bool = True):
    """
    This decorator is used for callbacks triggered by a button of name `btn_name`

    Args:
        btn_name: String with the name of the button.
        show_doc: If True, a Markdown pane with the doc string is returned the first time.
        show_shared_warning:
        show_exc: If True, a Markdown pane with the backtrace is returned if an exception is raised.
    """
    def decorator(func):
        @functools.wraps(func)
        def decorated(*args, **kwargs):
            self = args[0]
            btn = getattr(self, btn_name)
            if btn.clicks == 0:
                if show_doc and self._enable_show_doc:
                    doc = getattr(self, func.__name__).__doc__
                    if doc is None:
                        doc = f"\nNo docstring found for function `{func.__name__}`\n"
                    doc = textwrap.dedent(doc)
                    doc = f"## Description\n\n{doc}\n\n"
                    #print(doc)
                    objects = [my_md(doc)]
                    if show_shared_wdg_warning and self._enable_show_shared_wdg_warning:
                        warning = pn.pane.Alert(SHARED_WIDGETS_WARNING, alert_type="danger")
                        objects.append(warning)
                    return pn.Column(*objects, sizing_mode="stretch_width")
                else:
                    return None

            with ButtonContext(btn):
                return func(*args, **kwargs)

        f = pn.depends(f"{btn_name}.clicks")(decorated)
        if show_exc: f = show_exception(f)
        return f

    return decorator


def show_exception(func):
    """
    This decorator returns a Markdown pane with the backtrace
    if the function raises an exception.
    """
    @functools.wraps(func)
    def decorated(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as exc:
            s = traceback.format_exc()
            print(s)
            return pn.pane.Markdown(f"```shell\n{s}\n```", sizing_mode="stretch_both")

    return decorated


class HTMLwithClipboardBtn(pn.pane.HTML):
    """
    Receives an HTML string and returns an HTML pane with a button that allows users
    to copy the content to the system clipboard.
    Requires call to abipanel to load the JS extension.
    """

    # This counter is shared by all the instances.
    # We use it so that the js script is included only once.
    #_init_counter = [0]

    def __init__(self, object=None, btn_cls=None, **params):
        super().__init__(object=object, **params)

        #self._init_counter[0] += 1
        my_id = gen_id()
        btn_cls = "bk bk-btn bk-btn-default" if btn_cls is None else str(btn_cls)

        # Build new HTML string with js section if first call.
        new_text = f"""
<div id="{my_id}">{self.object}</div>
<br>
<button class="clip-btn {btn_cls}" type="button" data-clipboard-target="#{my_id}"> Copy to clipboard </button>
<hr>

"""
        if True: # self._init_counter[0] == 1:
            #new_text += " <script> $(document).ready(function() {new ClipboardJS('.clip-btn')}) </script> "
            # $(document).ready(function() {
            new_text += """ <script>
if (typeof abipy_clipboard === 'undefined') {
    var abipy_clipboard = new ClipboardJS('.clip-btn');
}

if (typeof abipy_notyf === 'undefined') {
    // Create an instance of Notyf
    var abipy_notyf = new Notyf();

    abipy_clipboard.on('success', function(e) {
        abipy_notyf.success('Text copied to clipboard');
    });

    abipy_clipboard.on('error', function(e) {
        abipy_notyf.error('Cannot copy text to clipboard');
    });
}
</script> """

        self.object = new_text


def mpl(fig, sizing_mode='stretch_width', with_controls=False, with_divider=True, **kwargs) -> pn.Column:
    """
    Helper function returning a panel Column with a matplotly pane followed by
    a divider and (optionally) controls to customize the figure.
    """
    col = pn.Column(sizing_mode=sizing_mode); ca = col.append

    #try:
    #    import ipympl
    #    has_ipympl = True
    #except ImportError:
    #    has_ipympl = False

    #if "interactive" not in kwargs and has_ipympl: # "ipympl" in sys.modules:
    #    print("mpl in interactive mode")
    #    kwargs["interactive"] = True

    if "tight" not in kwargs: kwargs["tight"] = True

    mpl_pane = pn.pane.Matplotlib(fig, **kwargs)
    ca(mpl_pane)

    if with_controls:
        ca(pn.Accordion(("matplotlib controls", mpl_pane.controls(jslink=True))))
        ca(pn.layout.Divider())

    if with_divider:
        ca(pn.layout.Divider())

    return col


def ply(fig, sizing_mode='stretch_both', with_chart_studio=False, with_help=False,
        with_divider=True, with_controls=False) -> pn.Column:
    """
    Helper function returning a panel Column with a plotly pane,  buttons to push the figure
    to plotly chart studio and, optionally, controls to customize the figure.
    """
    col = pn.Column(sizing_mode=sizing_mode); ca = col.append

    config = dict(
      responsive=True,
      #showEditInChartStudio=True,
      showLink=True,
      plotlyServerURL="https://chart-studio.plotly.com",
      )

    plotly_pane = pn.pane.Plotly(fig, config=config)
    ca(plotly_pane)

    if with_chart_studio:
        md = pn.pane.Markdown("""

The button on the left allows you to **upload the figure** to the
plotly [chart studio server](https://plotly.com/chart-studio-help/)
so that it is possible to share the figure or customize it via the chart studio editor.
In order to use this feature, you need to create a free account following the instructions
reported [in this page](https://plotly.com/chart-studio-help/how-to-sign-up-to-plotly/).
Then **add the following section** to your $HOME/.pmgrc.yaml configuration file:

```yaml
PLOTLY_USERNAME: john_doe  # Replace with your username
PLOTLY_API_KEY: secret     # To get your api_key go to: profile > settings > regenerate key
```

so that AbiPy can authenticate your user on the chart studio portal before pushing the figure to the cloud.
If everything is properly configured, a new window is automatically created in your browser.
""")

        btn = pnw.Button(name="Upload figure to chart studio server")

        def push_to_cs(event):
            with ButtonContext(btn):
                push_to_chart_studio(fig)

        btn.on_click(push_to_cs)

        if with_help:
            acc = pn.Accordion(("Help", md))
            ca(pn.Row(btn, acc))
        else:
            ca(pn.Row(btn))

    if with_controls:
        ca(pn.Accordion(("plotly controls", plotly_pane.controls(jslink=True))))

    if with_divider:
        ca(pn.layout.Divider())

    return col


def dfc(df: pd.DataFrame,
        #wdg_type: str = "dataframe",
        wdg_type: str ="tabulator",  # More recent version. Still problematic
        with_export_btn=True, with_controls=False, with_divider=True, transpose=False, **kwargs):
    """
    Helper function returning a panel Column with a DataFrame or Tabulator widget followed by
    a divider and (optionally) controls to customize the figure.

    Note that not all the options work as exected. See comments below.
    """
    if "disabled" not in kwargs: kwargs["disabled"] = True
    #if "sizing_mode" not in kwargs: kwargs["sizing_mode"] = "stretch_both"
    if "sizing_mode" not in kwargs: kwargs["sizing_mode"] = "scale_width"
    if transpose:
        df = df.transpose()

    if wdg_type == "dataframe":
        if "auto_edit" not in kwargs: kwargs["auto_edit"] = False
        w = pnw.DataFrame(df, **kwargs)
    elif wdg_type == "tabulator":
        w = pnw.Tabulator(df, **kwargs)
    else:
        raise ValueError(f"Don't know how to handle widget type: `{wdg_type}`")

    col = pn.Column(sizing_mode='stretch_both'); ca = col.append
    ca(w)

    if with_export_btn:
        # Define callbacks with closure.
        def to_xlsx():
            """
            Based on https://panel.holoviz.org/gallery/simple/file_download_examples.html
            NB: This requires xlsxwriter package else pandas raises ModuleNotFoundError.
            """
            output = io.BytesIO()
            writer = pd.ExcelWriter(output, engine='xlsxwriter')
            df.to_excel(writer, sheet_name="DataFrame")
            writer.save()  # Important!
            output.seek(0)
            return output

        def to_latex():
            """Convert DataFrame to latex string."""
            output = io.StringIO()
            df.to_latex(buf=output)
            output.seek(0)
            return output

        def to_md():
            """Convert DataFrame to markdown string."""
            output = io.StringIO()
            df.to_markdown(buf=output)
            output.seek(0)
            return output

        def to_json():
            """Convert DataFrame to json string."""
            output = io.StringIO()
            df.to_json(path_or_buf=output)
            output.seek(0)
            return output

        d = dict(
            xlsx=pnw.FileDownload(filename="data.xlsx", callback=to_xlsx),
            tex=pnw.FileDownload(filename="data.tex", callback=to_latex),
            md=pnw.FileDownload(filename="data.md", callback=to_md),
            json=pnw.FileDownload(filename="data.json", callback=to_json),
            )

        # For the time being we use a Row with buttons.
        #ca(pn.Row(*d.values(), sizing_mode="scale_width"))
        ca(pn.Card(*d.values(), title="Export table", collapsed=True,
                   sizing_mode='stretch_width', header_color="blue",
        ))

        #def download(event):
        #    file_download = d[event.new]
        #    #print(f'Clicked menu item: "{event.new}"')
        #    print(file_download)
        #    #file_download._clicks = -1
        #    file_download._transfer()
        #    return file_download.callback()

        # FIXME: Menu button occupies less space but the upload does not work
        #menu_btn = pnw.MenuButton(name='Export table', items=list(d.keys()))
        #menu_btn.on_click(download)
        #ca(menu_btn)

    if with_controls:
        ca(pn.Accordion(("dataframe controls", w.controls(jslink=True))))

    if with_divider:
        ca(pn.layout.Divider())

    return col


def my_md(string: str, **kwargs) -> pn.pane.Markdown:
    """
    Return a Markdown pane from `string`.
    Extra kwargs are passed to pane.Markdown.
    The string can contain links to Abinit variables in the wikilink format.

    .. example::

        The anaddb variable [[dipdip@anaddb]] has the same meaning as the Abinit variable [[dipdip]].
    """
    import re
    WIKILINK_RE = r'\[\[([^\[]+)\]\]'
    from abipy.abio.abivar_database.variables import get_codevars
    vars_code = get_codevars()

    def repl(match):
        var_name = match.group(1).strip()
        codename = "abinit"
        i = var_name.find("@")
        if i != -1:
            var_name, codename = var_name[:i], var_name[i+1:]
        try:
            var = vars_code[codename][var_name]
            return var.html_link()
        except:
            return f"WRONG LINK for {var_name}"

    string = re.sub(WIKILINK_RE, repl, string)

    # TODO: Latex
    extra_extensions = [
         "markdown.extensions.admonition",
         #'pymdownx.arithmatex',
         #'pymdownx.details',
         #"pymdownx.tabbed",
    ]

    md = pn.pane.Markdown(string, **kwargs)
    md.extensions.extend(extra_extensions)

    return md


class ButtonContext():
    """
    A context manager for buttons triggering computations on the server.

    This manager disables the button when we __enter__ and changes the name of the button to "running".
    It reverts to the initial state of the button once __exit__ is invoked, showing the Exception type
    in a "red" button if an exception is raised during the computation.

    This a very important tool because we need to disable the button when we start the computation
    to prevent the user from triggering multiple callbacks while the server is still working.
    At the same time, whathever happens in the callback, the button should go back to "clickable" mode
    when the callback returns so that the user can try to change the parameters and rerun.

    Note also that we want to provide some graphical feedback to the user if something goes wrong.
    At present we don't expose the python traceback on the client.
    It would be nice but we need panel machinery to do that.
    Moreover this is not the recommended approach for security reasons so we just change the "color"
    of the button and use the string representation of the exception as button name.
    """

    def __init__(self, btn: pnw.Button):
        self.btn = btn
        self.prev_name, self.prev_type = btn.name, btn.button_type

    def __enter__(self):
        # Disable the button.
        self.btn.name = "Running ..."
        self.btn.button_type = "warning"
        self.btn.loading = True
        self.btn.disabled = True
        return self.btn

    def __exit__(self, exc_type, exc_value, traceback):
        # First of all, reenable the button so that the user can stil interact with the GUI.
        self.btn.disabled = False
        self.btn.loading = False

        if exc_type:
            # Exception --> signal to the user that something went wrong for 2 seconds.
            self.btn.name = str(exc_type)
            self.btn.button_type = "danger"
            time.sleep(2)

        # Back to the original button state.
        self.btn.name, self.btn.button_type = self.prev_name, self.prev_type


class Loading():
    """
    A context manager for setting the loading attribute of a panel object.
    """

    def __init__(self, panel_obj, err_wdg=None, width=70):
        self.panel_obj = panel_obj
        self.err_wdg = err_wdg
        if err_wdg is not None: self.err_wdg.object = ""
        self.width = int(width)

    def __enter__(self):
        self.panel_obj.loading = True
        return self.panel_obj

    def __exit__(self, exc_type, exc_value, traceback):
        self.panel_obj.loading = False

        if self.err_wdg is not None:
            if exc_type:
                self.err_wdg.object = "```sh\n%s\n```" % textwrap.fill(str(exc_value), width=self.width)


class ActiveBar():
    """
    A context manager that sets progress.active to True on entry and False when we exit.
    """

    def __init__(self, progress, err_wdg=None, width=70):
        self.progress = progress
        self.err_wdg = err_wdg
        if err_wdg is not None: self.err_wdg.object = ""
        self.width = int(width)

    def __enter__(self):
        self.progress.active = True
        return self.progress

    def __exit__(self, exc_type, exc_value, traceback):
        self.progress.active = False

        if exc_type:
            # Change the color to signal the user that something bad happened.
            self.progress.bar_color = "danger"

            if self.err_wdg is not None:
                from textwrap import fill
                self.err_wdg.object = "```sh\n%s\n```" % fill(str(exc_value), width=self.width)


SHARED_WIDGETS_WARNING = """
Note some of the widgets are **shared by the different tabs**.
If you change the value of a variables in the active tab,
the same value will **automagically** appear in all the other tabs sharing the same
widget yet results/figures are not automatically recomputed."""

#In other words, if you change some variable in the active tab and then you move to another tab,
#the results/figures (if any) are stil computed with the **old input** hence you will have to
#recompute the new results by clicking the button."""


class AbipyParameterized(param.Parameterized):
    """
    Base class for AbiPy panels. Provides helper functions for typical operations needed for
    building dashboard and basic parameters supported by the subclasses.
    """

    verbose = param.Integer(0, bounds=(0, None), doc="Verbosity Level")
    mpi_procs = param.Integer(1, bounds=(1, None), doc="Number of MPI processes used for running Fortran code.")

    plotly_template = param.ObjectSelector(default="plotly",
                                           objects=["plotly", "plotly_white", "plotly_dark", "ggplot2",
                                                    "seaborn", "simple_white", "none"])

    # This flag is set to True if we are serving apps from the Abinit server.
    # It is used to impose limitations on what users can do and select the options that should be exposed.
    # For instance, structure_viewer == "Vesta" does not make sense in we are not serving from a local server.
    #
    has_remote_server = param.Boolean(False)

    warning = pn.pane.Markdown(SHARED_WIDGETS_WARNING, name="warning")

    def __init__(self, **params):

        super().__init__(**params)
        if self.has_remote_server:
            self.param.mpi_procs.bounds = (1, 1)
            print("Changing mpi_procs.bounds")
            print("self.param.mpi_procs:", self.param.mpi_procs.bounds)

        # This internal variables are used to enable/disable the output of the docstring and the warning message.
        self._enable_show_doc = True
        self._enable_show_shared_wdg_warning = True

    #def abiopen(self, filepath):
    #    from abipy.abilab import abiopen
    #    try:
    #        abiobj = abiopen(filepath)
    #        return abiobj
    #    finally:
    #        self._abiopened_files.append(filepath)

    def disable_show_messages(self, show_doc=False, show_shared_wdg_warning=False):
        """
        Context manager to disable the output of the docstring and the warning message.

        with self.disable_show_messages():
            pn.Row("## Callback with no info message", self.on_plot_phbands_and_phdos)

        """
        from abipy.tools.context_managers import temporary_change_attributes
        return temporary_change_attributes(self,
                                           _enable_show_doc=show_doc,
                                           _enable_show_shared_wdg_warning=show_shared_wdg_warning,
                                           )

    @pn.depends("plotly_template")
    def on_plotly_template(self):
        """
        Change the default plotly template.
        NB: It's not a good idea to expose this option when running the server as
        other users will be affected by this change hence this function is just for
        internal use.
        """
        import plotly.io as pio
        pio.templates.default = self.plotly_template

    @lazy_property
    def mpl_kwargs(self) -> dict:
        """Default arguments passed to AbiPy matplotlib plot methods."""
        return dict(show=False, fig_close=True)

    def pws_col(self, keys, **kwargs) -> pn.Column:
        return pn.Column(*self.pws(keys), **kwargs)

    def pws_row(self, keys, **kwargs) -> pn.Row:
        return pn.Row(*self.pws(keys), **kwargs)

    def wdg_box(self, keys, **kwargs) -> pn.WidgetBox:
        return pn.WidgetBox(*self.pws(keys), **kwargs)

    def pws(self, keys):
        """
        Helper function returning the list of parameters and widgets defined in self from a list of strings.
        Accepts also widget or parameter instances.
        """
        items, miss = [], []
        for k in keys:
            if isinstance(k, str):
                if k in self.param:
                    items.append(self.param[k])
                elif hasattr(self, k):
                    items.append(getattr(self, k))
                elif k.startswith("#"):
                    # Markdown string
                    items.append(k)
                    #items.append(pn.layout.Divider(margin=(-20, 0, 0, 0)))
                else:
                    miss.append(k)
            else:
                # Assume widget instance.
                items.append(k)

        #for item in items: print("item", item, "of type:", type(item))
        if miss:
            raise ValueError(f"Cannot find `{str(miss)}` in param or in attribute space")

        return items

    def get_summary_view_for_abiobj(self, abiobj, **kwargs):
        text = abiobj.to_string(verbose=self.verbose)

        view = pnw.Terminal(output=f"\n\n{text}",
            #height=1200, # Need this one else the terminal is not show properly
            sizing_mode='stretch_both',
        )
        #view = pn.Row(bkw.PreText(text=text, sizing_mode="scale_both"))
        return view

    def wdg_exts_with_get_panel(self, name='File extensions supported:'):
        """
        Return Select widget with the list of file extensions implementing a get_panel method.
        """
        from abipy.abilab import extcls_supporting_panel
        exts = [e[0] for e in extcls_supporting_panel(as_table=False)]
        return pnw.Select(name=name, options=exts)

    @staticmethod
    def html_with_clipboard_btn(html_str: str, **kwargs):
        if hasattr(html_str, "_repr_html_"):
            html_str = html_str._repr_html_()

        return HTMLwithClipboardBtn(html_str, **kwargs)

    @staticmethod
    def get_software_stack() -> pn.Column:
        """Return column with version of python packages in tabular format."""
        from abipy.abilab import software_stack
        return pn.Column("## Software stack:",
                         #pn.layout.Divider(margin=(-20, 0, 0, 0)),
                         dfc(software_stack(as_dataframe=True), with_export_btn=False),
                         sizing_mode="scale_width",
                         )

    @staticmethod
    def get_fileinput_section(file_input) -> pn.Column:
        # All credits go to:
        # https://github.com/MarcSkovMadsen/awesome-panel/blob/master/application/pages/styling/fileinput_area.py
        #
        css_style = """
        <style>
        .pnx-file-upload-area input[type=file] {
            width: 100%;
            height: 100%;
            border: 3px dashed #9E9E9E;
            background: transparent;
            border-radius: 5px;
            text-align: left;
            margin: auto;
        }
        </style>"""

        return pn.Column(pn.pane.HTML(css_style, width=0, height=0, sizing_mode="stretch_width", margin=0),
                         file_input, sizing_mode="stretch_width")

    @staticmethod
    def get_abifile_from_file_input(file_input, use_structure=False):
        #print("filename", file_input.filename, "\nvalue", file_input.value)
        workdir = tempfile.mkdtemp()

        fd, tmp_path = tempfile.mkstemp(suffix=file_input.filename)
        with open(tmp_path, "wb") as fh:
            fh.write(file_input.value)

        from abipy.abilab import abiopen
        abifile = abiopen(tmp_path)
        if use_structure:
            abifile = Structure.as_structure(abifile)
            # Remove the file since it's not needed anymore.
            shutil.rmtree(tmp_path, ignore_errors=True)

        return abifile

    def get_ebands_from_file_input(self, file_input, remove=True):
        """
        Read and return an |ElectronBands| object from a file_input widget.
        Return None if the file does not provide an ebands object.
        Remove the file if remove == True.
        """
        with self.get_abifile_from_file_input(file_input) as abifile:
            ebands = getattr(abifile, "ebands", None)
        if remove: abifile.remove()
        return ebands

    @staticmethod
    def get_alert_data_transfer() -> pn.pane.Alert:
        # https://discourse.holoviz.org/t/max-upload-size/2121/5
        return pn.pane.Alert("""
Please note that this web interface is not designed to handle **large data transfer**.
To post-process the data stored in a big file e.g. a WFK.nc file,
we strongly suggest executing the **abigui.py**  script on the same machine where the file is hosted.
Also, note that examples of post-processing scripts are available in the
[AbiPy gallery](https://abinit.github.io/abipy/gallery/index.html).

Last but not least, keep in mind that **the file extension matters** when uploading a file
so don't change the default extension used by ABINIT.
Also, use `.abi` for ABINIT input files and `.abo` for the main output file.
""", alert_type="info")

    @staticmethod
    def get_template_cls_from_name(template):
        return get_template_cls_from_name(template)

    def get_abinit_template_cls_kwds(self):
        return get_abinit_template_cls_kwds()

    def get_template_from_tabs(self, tabs, template, **tabs_kwargs):
        """
        This method receives panel Tabs or a dictionary,
        include them in a template and return the template instance.
        """
        if isinstance(tabs, dict):
            if "sizing_mode" not in tabs_kwargs: tabs_kwargs["sizing_mode"] = "stretch_width"
            tabs = pn.Tabs(*tabs.items(), **tabs_kwargs)

        if template is None or str(template) == "None":
            #tabs.append(return_to_top_html())
            return tabs

        cls = get_template_cls_from_name(template)
        #cls, kwargs = get_abinit_template_cls_kwds()

        kwargs = dict(
            # A title to show in the header. Also added to the document head meta settings and as the browser tab title.
            title=self.__class__.__name__,
            header_background="#ff8c00", # Dark orange
            favicon="/assets/img/abinit_favicon.ico",
            logo="/assets/img/abipy_logo.png", # TODO: Need new panel version to fix logo alignment in FastLIst.
            #sidebar_footer (str): Can be used to insert additional HTML. For example a menu, some additional info, links etc.
            #enable_theme_toggle=False,  # If True a switch to toggle the Theme is shown. Default is True.
        )

        template = cls(**kwargs)
        if hasattr(template.main, "append"):
            template.main.append(tabs)
            #template.main.append(return_to_top_html())
        else:
            # Assume main area acts like a GridSpec
            template.main[:,:] = tabs

        return template


def return_to_top_html():
    print("In return to top")

    #<a id="button"></a>

    html = r"""
<!--
Back to top button
https://codepen.io/matthewcain/pen/ZepbeR
-->

<style>
#button {
  display: inline-block;
  background-color: #FF9800;
  width: 50px;
  height: 50px;
  text-align: center;
  border-radius: 4px;
  position: fixed;
  bottom: 30px;
  right: 30px;
  transition: background-color .3s,
    opacity .5s, visibility .5s;
  opacity: 0;
  visibility: hidden;
  z-index: 1000;
}
#button::after {
  content: "\f077";
  font-family: FontAwesome;
  font-weight: normal;
  font-style: normal;
  font-size: 2em;
  line-height: 50px;
  color: #fff;
}
#button:hover {
  cursor: pointer;
  background-color: #333;
}
#button:active {
  background-color: #555;
}
#button.show {
  opacity: 1;
  visibility: visible;
}
</style>

<script>

// Code executed on page ready
$(function() {

    var btn = $('#button');
    console.log("button", btn);

    $(window).scroll(function() {
      if ($(window).scrollTop() > 300) {
        console.log("show");
        btn.addClass('btn show');
      } else {
        btn.removeClass('show');
        console.log("btn hide");
      }
    });

    btn.on('click', function(e) {
      e.preventDefault();
      $('html, body').animate({scrollTop:0}, '300');
    });

})
</script>
"""

    return pn.pane.HTML(html, width=0, height=0)


class PanelWithStructure(AbipyParameterized):
    """
    A paremeterized object with a |Structure| object.
    """

    structure_viewer = param.ObjectSelector(default="jsmol",
                                            objects=["jsmol", "vesta", "xcrysden", "vtk", "crystalk", "ngl",
                                                     "matplotlib", "plotly", "ase_atoms", "mayavi"])

    def __init__(self, structure: Structure, **params):

        super().__init__(**params)
        self.structure = structure

        if self.has_remote_server:
            # Change the list of allowed visualizers.
            self.param.structure_viewer.objects = ["jsmol", "crystalk", "ngl", "matplotlib", "plotly", "ase_atoms"]

        self.view_structure_btn = pnw.Button(name="View structure", button_type='primary')

    @depends_on_btn_click('view_structure_btn', show_shared_wdg_warning=False)
    def on_view_structure(self):
        """Visualize input structure."""
        v = self.structure_viewer

        if v == "jsmol":
            return jsmol_html(self.structure)

            #pn.extension(comms='ipywidgets') #, js_files=js_files)
            #view = self.structure.get_jsmol_view()
            #from ipywidgets_bokeh import IPyWidget
            #view = IPyWidget(widget=view) #, width=800, height=300)
            #import ipywidgets as ipw
            #from IPython.display import display
            #display(view)
            #return pn.Row(display(view))
            #return pn.ipywidget(view)
            #return pn.panel(view)
            #return pn.pane.IPyWidget(view)
            #print(view)
            #view = pn.Column(view, sizing_mode='stretch_width')
            #return view

        if v == "crystalk":
            view = self.structure.get_crystaltk_view()
            return pn.panel(view)

        if v == "plotly":
            return ply(self.structure.plotly(show=False))

        if v == "ngl":
            from pymatgen.io.babel import BabelMolAdaptor
            from pymatgen.io.xyz import XYZ
            # string_data = self.structure.to(fmt="xyz")

            #writer = BabelMolAdaptor(self)
            #string_data = str(XYZ(self.structure))
            #adapt = BabelMolAdaptor.from_string(string_data, file_format="xyz")
            ##pdb_string =
            #print(pdb_string)

            #from awesome_panel_extesions.pane.widgets.ngl_viewer import NGLViewer
            #view = NGLViewer()

            #view.pdb_string = pdb_string
            return view

            #js_files = {'ngl': 'https://cdn.jsdelivr.net/gh/arose/ngl@v2.0.0-dev.33/dist/ngl.js'}
            #pn.extension(comms='ipywidgets', js_files=js_files)
            #view = self.structure.get_ngl_view()
            #return pn.panel(view)

            #pn.config.js_files["ngl"]="https://cdn.jsdelivr.net/gh/arose/ngl@v2.0.0-dev.33/dist/ngl.js"
            #pn.extension()

            html = """<div id="viewport" style="width:100%; height:100%;"></div>
            <script>
            stage = new NGL.Stage("viewport");
            stage.loadFile("rcsb://1NKT.mmtf", {defaultRepresentation: true});
            </script>"""

            ngl_pane = pn.pane.HTML(html, height=500, width=500)
            return pn.Row(ngl_pane)
            view = self.structure.get_ngl_view()

        #return self.structure.crystaltoolkitview()
        #import nglview as nv
        #view = nv.demo(gui=False)

        if v == "ase_atoms":
            return mpl(self.structure.plot_atoms(rotations="default", **self.mpl_kwargs))

        return self.structure.visualize(appname=self.structure_viewer)

    def get_structure_view(self) -> pn.Row:
        """
        Return Row with widgets to visualize the structure.
        """
        return pn.Row(
            self.pws_col(["## Visualize structure",
                          "structure_viewer",
                          "view_structure_btn",
                          ]),
            pn.Column(self.on_view_structure, self.get_structure_info())
        )

    def get_structure_info(self) -> pn.Column:
        """
        Return Column with lattice parameters, angles and atomic positions grouped by type.
        """
        return get_structure_info(self.structure)


def get_structure_info(structure: Structure) -> pn.Column:
    """
    Return Column with lattice parameters, angles and atomic positions grouped by type.
    """
    col = pn.Column(sizing_mode='scale_width'); ca = col.append; cext = col.extend

    d = structure.get_dict4pandas(with_spglib=True)

    keys = index = [#"formula", "natom", "volume",
                    "abi_spg_number",
                    "spglib_symb", "spglib_num",  "spglib_lattice_type"]
    df_spg = pd.Series(data=d, index=index).to_frame()
    cext(["# Spacegroup:", dfc(df_spg, with_export_btn=False)])

    # Build dataframe with lattice lenghts.
    rows = []; keys = ("a", "b", "c")
    rows.append({k: d[k] * abu.Ang_Bohr for k in keys})
    rows.append({k: d[k] for k in keys})
    df_len = pd.DataFrame(rows, index=["Å", "Bohr"]).transpose().rename_axis("Lattice lenghts")

    # Build dataframe with lattice angles.
    rows = []; keys = ("alpha", "beta", "gamma")
    rows.append({k: d[k] for k in keys})
    rows.append({k: np.radians(d[k]) for k in keys})
    df_ang = pd.DataFrame(rows, index=["Degrees", "Radians"]).transpose().rename_axis("Lattice angles")

    cext(["# Lattice lengths:", dfc(df_len, with_export_btn=False)])
    cext(["# Lattice angles:", dfc(df_ang, with_export_btn=False)])
    #row = pn.Row(dfc(df_len, with_export_btn=False), dfc(df_ang, with_export_btn=False), sizing_mode="scale_width")
    #ca(row)

    # Build dataframe with atomic positions grouped by element symbol.
    symb2df = structure.get_symb2coords_dataframe()
    accord = pn.Accordion(sizing_mode='stretch_width')
    for symb, df in symb2df.items():
        accord.append((f"Coordinates of {symb} sites:", dfc(df, with_export_btn=False)))
    ca(accord)

    return col


class NcFileViewer(AbipyParameterized):
    """
    This class implements toool to inspect dimensions and variables stored in a netcdf file.

    Relyes on the API provided by `AbinitNcFile` defined in core.mixins.py
    """

    nc_path = param.String("/", doc="nc group")

    def __init__(self, ncfile, **params):
        super().__init__(**params)
        self.ncfile = ncfile
        self.netcdf_info_btn = pnw.Button(name="Show info", button_type='primary')

    def get_ncfile_view(self) -> pn.Column:
        return pn.Column(
                self.netcdf_info_btn,
                self.on_netcdf_info_btn,
                sizing_mode='stretch_width',
        )

    @depends_on_btn_click('netcdf_info_btn')
    def on_netcdf_info_btn(self) -> pn.Column:
        """
        This Tab allows one to
        """
        # TODO: Finalize the implementation.
        col = pn.Column(sizing_mode='stretch_width'); ca = col.append

        #nc_grpname = pnw.Select(name="nc group name", options=["/"])
        input_string = self.ncfile.get_input_string()
        ca(f"## Input String")
        ca(bkw.PreText(text=input_string))

        #ca(f"## Global attributes")

        # Get dataframe with dimensions.
        dims_df = self.ncfile.get_dims_dataframe(path=self.nc_path)
        ca(f"## Dimensions in nc group: {self.nc_path}")
        ca(dfc(dims_df))
        #ca(f"## Variables")

        return col


class PanelWithElectronBands(PanelWithStructure):
    """
    Provide widgets and views for operating on |ElectronBands| object.
    """

    # Bands plot
    with_gaps = param.Boolean(False)
    with_kpoints_plot = param.Boolean(False)

    #ebands_ylims
    #ebands_e0
    # e0: Option used to define the zero of energy in the band structure plot. Possible values:
    #     - `fermie`: shift all eigenvalues to have zero energy at the Fermi energy (`self.fermie`).
    #     -  Number e.g e0=0.5: shift all eigenvalues to have zero energy at 0.5 eV
    #     -  None: Don't shift energies, equivalent to e0=0
    set_fermie_to_vbm = param.Boolean(False, label="Set Fermi energy to VBM")

    # e-DOS plot.
    edos_method = param.ObjectSelector(default="gaussian", label="Integration method for e-DOS",
                                       objects=["gaussian", "tetra"])
    edos_step_ev = param.Number(0.1, bounds=(1e-6, None), step=0.1, label='e-DOS step in eV')
    edos_width_ev = param.Number(0.2, step=0.05, bounds=(1e-6, None), label='e-DOS Gaussian broadening in eV')

    # SKW interpolation of the KS band energies (Abipy version).
    skw_lpratio = param.Integer(5, bounds=(1, None), label="SKW lpratio")
    skw_line_density = param.Integer(20, label="SKW line density")
    skw_ebands_kpath = None
    skw_ebands_kpath_fileinput = param.FileSelector(path="*.nc")

    # Fermi surface.
    fs_viewer = param.ObjectSelector(default="matplotlib", objects=["matplotlib", "xcrysden"])

    # Ifermi UI.
    ifermi_wigner_seitz = param.Boolean(True, label="Use Wigner Seitz cell",
                                        doc="Controls whether the cell is the Wigner-Seitz cell" +
                                            "or the reciprocal unit cell parallelepiped.")
    ifermi_interpolation_factor = param.Integer(default=8, label="Interpolation factor", bounds=(1, None),
                                                doc="The factor by which the band structure will be interpolated.")

    ifermi_eref = param.ObjectSelector(default="fermie", label="Energy reference",
                                       objects=["fermie", "cbm", "vbm"])

    ifermi_with_velocities = param.Boolean(True, label="Compute group velocities",
            doc="Generate the Fermi surface and calculate the group velocity at the center of each triangular face")
    ifermi_offset_eV = param.Number(default=0.0, label="Energy offset (eV) from energy reference",
                                    doc="Energy offset from the Fermi energy at which the isosurface is calculated.")
    ifermi_plot_type = param.ObjectSelector(default="plotly", label="Plot type", objects=["plotly", "matplotlib"])

    # These are used to implement plots in which we need to upload an additional file
    # For instance bands + edos.
    # For the max size of file see: https://github.com/holoviz/panel/issues/1559
    ebands_kpath = None
    ebands_kpath_fileinput = param.FileSelector(path="*.nc")
    ebands_kmesh = None
    ebands_kmesh_fileinput = param.FileSelector(path=".nc")

    effmass_accuracy = param.Integer(default=4, bounds=(1, None), label="Finite difference accuracy")
    effmass_degtol_ev = param.Number(default=1e-3, bounds=(0, None), label="Window in eV above/below the CBM/VBM")
    effmass_spin = param.ObjectSelector(default=0, objects=[0, 1], label="Spin index")

    def __init__(self, ebands, **params):

        self.ebands = ebands
        PanelWithStructure.__init__(self, structure=ebands.structure, **params)

        # Create buttons
        self.plot_ebands_btn = pnw.Button(name="Plot e-bands", button_type='primary')
        self.plot_edos_btn = pnw.Button(name="Plot e-DOS", button_type='primary')
        self.plot_skw_btn = pnw.Button(name="Plot SKW interpolant", button_type='primary')

        # Fermi surface plotter.
        #objects = [None, "matplotlib", "xcrysden"]
        #if self.has_remote_server: objects = [None, "matplotlib"]
        #self.plot_fs_viewer_btn = pnw.Button(name="Plot SKW interpolant", button_type='primary')

        self.plot_ifermi_btn = pnw.Button(name="Plot Fermi surface", button_type='primary')
        #self.ifermi_plane_normal = pnw.LiteralInput(name='Plane normal (list)', value=[0, 0, 0], type=list,
        #                                            placeholder="Enter normal in reduced coordinates")
        #self.ifermi_distance = pn.widgets.RangeSlider(
        #        name='distance', start=0, end=2 * max(ebands.structure.reciprocal_lattice.abc),
        #        value=(0, 0), step=0.01)

        #ebands_kpath_fileinput = pnw.FileInput(accept=".nc")
        #ebands_kmesh_fileinput = pnw.FileInput(accept=".nc")

        self.plot_effmass_btn = pnw.Button(name="Plot effective masses", button_type='primary')
        if ebands.nsppol != 2:
            self.param.effmass_spin.objects = [0]

    @staticmethod
    def _get_ebands_from_bstring(bstring):
        from abipy.electrons import ElectronBands
        return ElectronBands.from_binary_string(bstring)

    @pn.depends("ebands_kpath_fileinput", watch=True)
    def get_ebands_kpath(self):
        """
        Receives the netcdf file selected by the user as binary string.
        """
        self.ebands_kpath = self._get_ebands_from_bstring(self.ebands_kpath_fileinput)

    @pn.depends("ebands_kmesh_fileinput", watch=True)
    def get_ebands_kmesh(self):
        """
        Receives the netcdf file selected by the user as binary string.
        """
        self.ebands_kmesh = self._get_ebands_from_bstring(self.ebands_kmesh_fileinput)

    @pn.depends("skw_ebands_kpath_fileinput", watch=True)
    def get_skw_ebands_kpath(self):
        """
        Receives the netcdf file selected by the user as binary string.
        """
        self.skw_ebands_kpath = self._get_ebands_from_bstring(self.skw_ebands_kpath_fileinput)

    def get_plot_ebands_view(self) -> pn.Row:
        return pn.Row(
            self.pws_col(["### e-Bands Plot Options",
                          "with_gaps", "set_fermie_to_vbm", "with_kpoints_plot", "plot_ebands_btn",
                         ]),
            self.on_plot_ebands_btn
        )

    @depends_on_btn_click('plot_ebands_btn')
    def on_plot_ebands_btn(self) -> pn.Column:
        """
        This Tab allows one to plot the KS energies stored in the netcdf file
        as well as the associated list of **k**-points in the Brillouin.
        """
        if self.set_fermie_to_vbm:
            self.ebands.set_fermie_to_vbm()

        sz_mode = "stretch_width"
        col = pn.Column(sizing_mode=sz_mode); ca = col.append
        ca("## Electronic band structure:")
        fig1 = self.ebands.plotly(e0="fermie", ylims=None, with_gaps=self.with_gaps, max_phfreq=None, show=False)
        ca(ply(fig1))

        nkpt = len(self.ebands.kpoints)
        ktype = "IBZ sampling" if self.ebands.kpoints.is_ibz else "**k**-path"
        max_nkpt = 2000

        if self.with_kpoints_plot:
            ca(f"## Brillouin zone and {ktype}:")
            if nkpt < max_nkpt:
                kpath_pane = ply(self.ebands.kpoints.plotly(show=False), with_divider=False)
                df_kpts = dfc(self.ebands.kpoints.get_highsym_datataframe(), with_divider=False)
                ca(pn.Row(kpath_pane, df_kpts, sizing_mode=sz_mode))
            else:
                ca(f"k-points won't be shown as nkpt: {nkpt} is greater than {max_nkpt}")
        ca(pn.layout.Divider())

        return col

    def get_plot_edos_view(self) -> pn.Row:
        return pn.Row(
                self.pws_col(["## E-DOS Options", "edos_method", "edos_step_ev",
                              "edos_width_ev", "plot_edos_btn"]),
                self.on_plot_edos_btn
                )

    @depends_on_btn_click('plot_edos_btn')
    def on_plot_edos_btn(self) -> pn.Row:
        """
        Button triggering edos plot.
        """
        edos = self.ebands.get_edos(method=self.edos_method, step=self.edos_step_ev, width=self.edos_width_ev)

        return pn.Row(ply(edos.plotly(show=False)), sizing_mode='scale_width')

    def get_skw_view(self) -> pn.Row:
        """
        Column with widgets to use SKW.
        """
        wdg = pn.Param(
            self.param['skw_ebands_kpath_fileinput'],
            widgets={'skw_ebands_kpath_fileinput': pn.widgets.FileInput}
        )

        return pn.Row(
            self.pws_col(["## SKW options",
                          "skw_lpratio", "skw_line_density", "with_gaps",
                          "## Upload GSR.nc file with *ab-initio* energies along a k-path to compare with", wdg,
                          "plot_skw_btn",
                          ]),
            self.on_plot_skw_btn)

    @depends_on_btn_click('plot_skw_btn')
    def on_plot_skw_btn(self) -> pn.Column:
        """
        Button triggering SKW plot.
        """
        col = pn.Column(sizing_mode='stretch_width'); ca = col.append

        intp = self.ebands.interpolate(lpratio=self.skw_lpratio, line_density=self.skw_line_density,
                                       kmesh=None, is_shift=None, bstart=0, bstop=None, filter_params=None,
                                       verbose=self.verbose)

        ca("## SKW interpolated bands along an automatically selected high-symmetry **k**-path")
        ca(ply(intp.ebands_kpath.plotly(with_gaps=self.with_gaps, show=False)))

        if self.skw_ebands_kpath is not None:
            ca("## Input bands taken from file uploaded by user:")
            ca(ply(self.skw_ebands_kpath.plotly(with_gaps=self.with_gaps, show=False)))

            # Use line_density 0 to interpolate on the same set of k-points given in self.skw_ebands_kpath
            vertices_names = []
            for kpt in self.skw_ebands_kpath.kpoints:
                vertices_names.append((kpt.frac_coords, kpt.name))

            intp = self.ebands.interpolate(lpratio=self.skw_lpratio, vertices_names=vertices_names, line_density=0,
                                            kmesh=None, is_shift=None, bstart=0, bstop=None, filter_params=None,
                                            verbose=self.verbose)

            plotter = self.skw_ebands_kpath.get_plotter_with("Input", "Interpolated", intp.ebands_kpath)
            ca("## Input bands vs SKW interpolated bands:")
            ca(ply(plotter.combiplotly(show=False)))

        return col

    def get_effmass_view(self) -> pn.Row:
        """
        Return Row with widgets to compute effective masses with finite diff.
        """
        return pn.Row(
            self.pws_col(["### Effective masses options",
                          "effmass_accuracy",
                          "effmass_degtol_ev",
                          "effmass_spin",
                          "plot_effmass_btn",
                         ]),
            self.on_plot_effmass_btn
        )

    @depends_on_btn_click('plot_effmass_btn')
    def on_plot_effmass_btn(self) -> pn.Column:
        """
        Compute and visualize effective masses with finite differences.

        Note that the quality of the results strongly depend on the step i.e.
        the separation between two consecutive points along the k-path.
        The accuracy option allows one to change the number of points for the finite difference.
        """
        emana = self.ebands.get_effmass_analyzer()
        acc = self.effmass_accuracy
        degtol_ev = self.effmass_degtol_ev
        spin = self.effmass_spin

        col = pn.Column(sizing_mode='stretch_width'); ca = col.append

        if emana.select_vbm():
            ca(f"## Effective masses at the VBM with accuracy {acc}:")
            for segment in emana.segments:
                ca(mpl(segment.plot_emass(acc=acc, spin=spin, degtol_ev=degtol_ev, show=False)))
                ca("### effmass wrt accuracy and step: %.3f Ang-1" % segment.dk)
                df = segment.get_dataframe_with_accuracies()
                ca(dfc(df, with_export_btn=False))

        if emana.select_cbm():
            ca(f"## Effective masses at the CBM with accuracy {acc}:")
            for segment in emana.segments:
                ca(mpl(segment.plot_emass(acc=acc, spin=spin, degtol_ev=degtol_ev, show=False)))
                ca("### effmass wrt accuracy and step: %.3f Ang-1" % segment.dk)
                df = segment.get_dataframe_with_accuracies()
                ca(dfc(df, with_export_btn=False))

        return col

    def get_ifermi_view(self):
        """
        Widgets to visualize the Fermi surface with ifermi package.
        """
        controls = self.pws_col([
                          "ifermi_offset_eV", "ifermi_eref", "ifermi_wigner_seitz", "ifermi_interpolation_factor",
                          "ifermi_with_velocities", # "ifermi_plane_normal", "ifermi_distance",
                          "ifermi_plot_type", "plot_ifermi_btn",
                          ])

        return pn.Row(pn.Column("## ifermi options", controls), self.on_plot_ifermi_btn)

    @depends_on_btn_click('plot_ifermi_btn')
    def on_plot_ifermi_btn(self):
        """
        This Tab allows you to interpolate KS energies defined in the IBZ
        and visualize isosurfaces in the BZ using the [ifermi](https://fermisurfaces.github.io/IFermi/) package.

        The energy of the isosurface is given by: `energy_reference` + `energy_offset`.

        * use `fermie` with zero offset to visualize the Fermi surface in metals.
        * for electron pockets in semiconductors, use `cbm` with a positive offset e.g. 0.1 eV.
        * for hole pockets in semiconductors, use `vbm` and a  negative offset e.g. -0.1 eV.
        """
        # interpolate the energies onto a dense k-point mesh
        r = self.ebands.get_ifermi_fs(interpolation_factor=self.ifermi_interpolation_factor,
                                      mu=self.ifermi_offset_eV,
                                      eref=self.ifermi_eref,
                                      wigner_seitz=self.ifermi_wigner_seitz,
                                      calculate_dimensionality=False,
                                      with_velocities=self.ifermi_with_velocities,
                                      )

        plt = r.fs_plotter.get_plot(plot_type=self.ifermi_plot_type)

        if self.ifermi_plot_type == "plotly":
            fig = ply(plt, sizing_mode="stretch_both")
        elif self.ifermi_plot_type == "matplotlib":
            fig = plt.gcf()
            plt.close(fig=fig)
            fig = mpl(fig)
        else:
            raise NotImplementedError(f"{self.ifermi_plot_type}")

        col = pn.Column(sizing_mode="stretch_width")
        ca = col.append
        if self.ifermi_wigner_seitz:
            ca("## Energy isosurface in the Wigner-Seitz unit cell")
        else:
            ca("## Energy isosurface in the reciprocal unit cell parallelepiped")
        ca(fig)

        ene_range = [-3, 3]
        fig = self.ebands.boxplotly(e0=r.abs_isoenergy, ene_range=ene_range, show=False)
        ca(f"""
## Energy boxplot:

- Energy zero set at the absolute isoenergy: {r.abs_isoenergy:.3f} (eV)
- Energy range around zero: {ene_range} (eV)
""")
        ca(ply(fig, sizing_mode="stretch_both"))

        # TODO: This requires more testing
        #ifermi_plane_normal = self.ifermi_plane_normal.value
        #if any(c != 0 for c in ifermi_plane_normal):
        #    print("fs_plane normal:", ifermi_plane_normal)
        #    print("abc reciprocal", self.ebands.structure.reciprocal_lattice.abc)
        #    from ifermi.plot import FermiSlicePlotter
        #    for distance in self.ifermi_distance.value:
        #        # generate Fermi slice along the (0 0 1) plane going through the Γ-point.
        #        ca(f"## Fermi slice along the {ifermi_plane_normal} plane going through the Γ-point at distance: {distance}")
        #        fermi_slice = r.fs.get_fermi_slice(plane_normal=ifermi_plane_normal, distance=distance)
        #        slice_plotter = FermiSlicePlotter(fermi_slice)
        #        plt = slice_plotter.get_plot()
        #        fig = plt.gcf()
        #        plt.close(fig=fig)
        #        fig = mpl(fig)
        #        ca(fig)

        #ca(pn.layout.Divider())
        #ca("## Powered by [ifermi](https://fermisurfaces.github.io/IFermi/)")

        return col

    #@depends_on_btn_click('plot_fs_viewer_btn')
    #def on_plot_fs_viewer_btn(self):

    #    # Get eb3d (memoized)
    #    eb3d = self._eb3d = self.ebands.get_ebands3d()

    #    if self.fs_viewer == "matplotlib":
    #        # Use matplotlib to plot isosurfaces corresponding to the Fermi level (default)
    #        # Warning: requires skimage package, rendering could be slow.
    #        fig = eb3d.plot_isosurfaces(e0="fermie", cmap=None, **self.mpl_kwargs)
    #        return pn.Row(mpl(fig), sizing_mode='scale_width')

    #    elif self.fs_viewer == "xcrysden":
    #       # Alternatively, it's possible to export the data in xcrysden format
    #       # and then use `xcrysden --bxsf mgb2.bxsf`
    #       #eb3d.to_bxsf("mgb2.bxsf")
    #       # If you have mayavi installed, try:
    #       #eb3d.mvplot_isosurfaces()
    #       raise NotImplementedError(f"Invalid choice for fs_viewer: {self.fs_viewer}")

    #    else:
    #        raise ValueError(f"Invalid choice for fs_viewer: {self.fs_viewer}")

    #def get_fsviewer_view(self):
    #    """
    #    Widgets to visualize the Fermi surface with ifermi
    #    """
    #    #return pn.Row(pn.Column("## FS Viewer options", self.fs_viewer, plot_fs_viewer_btn),
    #    #                        self.on_plot_fs_viewer_btn)

    #    return pn.Row(
    #        self.pws_col(["## FS Viewer options",
    #                      "fs_viewer", "plot_fs_viewer_btn"
    #                      ]),
    #        self.on_plot_fs_viewer_btn)


class BaseRobotPanel(AbipyParameterized):
    """
    Base class for panels with AbiPy robot.
    """

    def __init__(self, robot, **params):
        self.robot = robot
        self.compare_params_btn = pnw.Button(name="Compare structures", button_type='primary')
        self.transpose_params = pnw.Checkbox(name='Transpose table', default=True)

        super().__init__(**params)

    @depends_on_btn_click("compare_params_btn")
    def on_compare_params_btn(self):
        """
        Compare lattice parameters and atomic positions.
        """
        col = pn.Column(sizing_mode='stretch_width'); ca = col.append
        transpose = self.transpose_params.value

        dfs = self.robot.get_structure_dataframes()
        ca("# Lattice dataframe")
        ca(dfc(dfs.lattice, transpose=transpose))

        ca("# Parameters dataframe")
        ca(dfc(self.robot.get_params_dataframe(), transpose=transpose))

        accord = pn.Accordion(sizing_mode='stretch_width')
        accord.append(("Atomic positions", dfc(dfs.coords, transpose=transpose)))
        ca(accord)

        return col

    # TODO: widgets to change robot labels.
    def get_compare_params_widgets(self):
        """
        """
        return pn.Row(pn.Column(
            self.compare_params_btn, self.transpose_params),
            self.on_compare_params_btn,
            sizing_mode="scale_both")


class PanelWithEbandsRobot(BaseRobotPanel):
    """
    Mixin class for panels with a robot that owns a list of of |ElectronBands|.
    """

    def __init__(self, robot, **params):

        BaseRobotPanel.__init__(self, robot=robot, **params)

        # Widgets to plot ebands.
        self.ebands_plotter_mode = pnw.Select(name="Plot Mode", value="gridplot",
                                              options=["gridplot", "combiplot", "boxplot", "combiboxplot"]) # "animate",
        self.ebands_plotter_btn = pnw.Button(name="Plot", button_type='primary')
        self.ebands_df_checkbox = pnw.Checkbox(name='With Ebands DataFrame', value=False)

        # Widgets to plot edos.
        self.edos_plotter_mode = pnw.Select(name="Plot Mode", value="gridplot", options=["gridplot", "combiplot"])
        self.edos_plotter_btn = pnw.Button(name="Plot", button_type='primary')

    def get_ebands_plotter_widgets(self) -> pn.Column:
        return pn.Column(self.ebands_plotter_mode, self.ebands_df_checkbox, self.ebands_plotter_btn)

    @depends_on_btn_click("ebands_plotter_btn")
    def on_ebands_plotter_btn(self) -> pn.Row:
        """
        Plot the electronic density of states.
        """
        ebands_plotter = self.robot.get_ebands_plotter()
        plot_mode = self.ebands_plotter_mode.value
        plot_func = getattr(ebands_plotter, plot_mode, None)
        if plot_func is None:
            raise ValueError("Don't know how to handle plot_mode: %s" % plot_mode)

        fig = plot_func(**self.mpl_kwargs)
        col = pn.Column(mpl(fig), sizing_mode='scale_width')

        if self.ebands_df_checkbox.value:
            df = ebands_plotter.get_ebands_frame(with_spglib=True)
            col.append(dfc(df))

        return pn.Row(col, sizing_mode='scale_width')

    def get_edos_plotter_widgets(self) -> pn.Column:
        return pn.Column(self.edos_plotter_mode, self.edos_plotter_btn)

    @depends_on_btn_click("edos_plotter_btn")
    def on_edos_plotter_btn(self) -> pn.Row:
        """
        Plot the electronic density of states.
        """
        edos_plotter = self.robot.get_edos_plotter()
        plot_mode = self.edos_plotter_mode.value
        plot_func = getattr(edos_plotter, plot_mode, None)
        if plot_func is None:
            raise ValueError("Don't know how to handle plot_mode: %s" % plot_mode)

        fig = plot_func(**self.mpl_kwargs)

        return pn.Row(pn.Column(mpl(fig)), sizing_mode='scale_width')


def jsmol_html(structure, supercell=(1, 1, 1), width=700, height=700, color="black", spin="false")  -> pn.Column:

    cif_str = structure.write_cif_with_spglib_symms(None, ret_string=True) #, symprec=symprec

    # There's a bug in boken when we use strings with several '" quotation marks
    # To bypass the problem I create a json list of strings and then I use a js variable
    # to recreate the cif file by joining the tokens with: var string = elements.join('\n');
    import json
    lines = cif_str.split("\n")
    lines = json.dumps(lines, indent=4)
    #print("lines:", lines)

    jsmol_div_id = gen_id()
    jsmol_app_name = "js1"
    supercell = "{" + ",".join(str(s) for s in supercell) + "}"
    #supercell = "{2, 2, 2}"

    # http://wiki.jmol.org/index.php/Jmol_JavaScript_Object/Functions#getAppletHtml

    html = f"""
<script type="text/javascript">
    $(document).ready(function() {{

       const lines = {lines};
       var script_str = 'load inline " ' + lines.join('\\n') + '" {supercell}';
       //console.log(script_str);

       var Info = {{
           color: "{color}",
           spin: {spin},
           antialiasDisplay: true,
           width: {width},
           height: {height},
           j2sPath: "https://chemapps.stolaf.edu/jmol/jsmol/j2s",
           serverURL: "https://chemapps.stolaf.edu/jmol/jsmol/php/jsmol.php",
           script: script_str,
           use: 'html5',
           disableInitialConsole: true,
           disableJ2SLoadMonitor: true,
           debug: false
       }};

       $("#{jsmol_div_id}").html(Jmol.getAppletHtml("{jsmol_app_name}", Info));
    }});
</script>

<div id="{jsmol_div_id}" style="height: 100%; width: 100%; position: relative;"></div>
"""

    #print(html)
    return pn.Column(pn.pane.HTML(html, sizing_mode="stretch_width"), sizing_mode="stretch_width")



#!/usr/bin/env python
r"""
Plot convergence
================

This example shows how to plot the convergence of
scalar quantities wrt to one independent variable.
For instance, the convergence of the energy per atom
with the cutoff energy ecut.
"""

#%%
# Let's assume we have the total energy per atom
# and the pressure in GPa as a function of ecut
# stored in three python lists:

ecut_list = [35., 40., 45., 50., 55., 60., 65.]

energy_per_atom_ev = [
-444.29611271, -444.3113239, -444.3124781,
-444.31251066, -444.31254606, -444.31256315, -444.31258609,
]

pressure_gpa =  [
-17.56316746, -16.00634717, -15.80295881,
-15.79975677, -15.79858973, -15.79811464, -15.79929941,
]

#%%
# To plot energy vs ecut with a convergence window of 1e-3
# centered around our best estimate i.e. the value obtained
# with the highest ecut, use the high-level API `from_xy_label_vals`:

from abipy.tools.plotting import ConvergenceAnalyzer

ca = ConvergenceAnalyzer.from_xy_label_vals("ecut (Ha)", ecut_list,
                                            "E/natom (eV)", energy_per_atom_ev, tols=1e-3)

print(ca)
ca.plot()

#%%
# To analyze the convergence of two or more quantities, use the low-level API
# that expects two dictionaries mapping the y-label used in the plot to
# (1) the list of values and (2) the associated absolute tolerance as in:

yvals_dict = {
    "E/natom (eV)": energy_per_atom_ev,
    "P (GPa)": pressure_gpa,
}

ytols_dict = {
    "E/natom (eV)": 1e-3,
    "P (GPa)": 1e-1,
}

ca = ConvergenceAnalyzer("ecut (Ha)", ecut_list, yvals_dict, ytols_dict)

#print(ca)
#ca.plot()

#%%
# To specify multiple convergence criteria, replace scalars with tuples as in:

ytols_dict = {
    "E/natom (eV)": (1e-3, 1e-4),
    "P (GPa)": (1e-1, 1e-2),
}

ca = ConvergenceAnalyzer("ecut (Ha)", ecut_list, yvals_dict, ytols_dict)
#print(ca)
ca.plot()


#%%
# Finally, if you have a (xlsf or csv) file with columns named e.g.: `ecut`, `energy_per_atom`, `pressure`
# and you want to plot the results with ConvergenceAnalyzer, use the `from_file` method
# with a `ytols_dict` whose keys define the columns in tabular data:

ytols_dict = {
    "ecut": (1e-3, 1e-4),
    "pressure": (1e-1, 1e-2),
}
#ca = ConvergenceAnalyzer.from_file(filepath, ytols_dict)

# To change the x/y-labels in the plot, use:
#ca.set_label("ecut", "ecut (Ha)")
#ca.set_label("energy_per_atom", "E/natom (eV)")
#ca.set_label("pressure", "P (GPa)")

#print(ca)
#ca.plot()



""""Panels to interact with AbiPy flows."""
from __future__ import annotations

import param
import panel as pn
import panel.widgets as pnw
#import bokeh.models.widgets as bkw

from panel.viewable import Viewer
from abipy.panels.core import mpl, ply, dfc, depends_on_btn_click
from abipy.panels.nodes import NodeParameterized
from abipy.flowtk.tasks import AbinitTask
from abipy.flowtk.flows import Flow


class WorkTaskSelector(Viewer):

    task = param.ClassSelector(class_=AbinitTask, doc="Task object")

    def __init__(self, flow: Flow, **params):
        self._wstr2work = {f"w{i} ({work.__class__.__name__}, len: {len(work)})": work
                           for (i, work) in enumerate(flow.works)}
        options = list(self._wstr2work.keys())
        self.work_select = pnw.Select(name="Select a Work", value=options[0], options=options)

        options = [f"t{i} ({task.__class__.__name__}, {str(task.status)})"
                   for (i, task) in enumerate(flow[0])]
        self.task_select = pnw.Select(name="Select a Task in the Work", value=options[0], options=options)

        super().__init__(**params)

        self.layout = pn.Column(self.work_select, self.task_select)
        self.sync_widgets()

    def __panel__(self):
        return self.layout

    @pn.depends('work_select.value', watch=True)
    def update_work(self) -> None:
        self.work = self._wstr2work[self.work_select.value]
        self.task_select.options = [f"t{i} ({task.__class__.__name__}, {str(task.status)})"
                                    for (i, task) in enumerate(self.work)]
        self.task = self.work[0]

    @pn.depends('task_select.value', watch=True)
    def sync_widgets(self) -> None:
        self.work = self._wstr2work[self.work_select.value]
        task_idx = int(self.task_select.value[1:].split()[0])
        self.task = self.work[task_idx]


class FlowPanel(NodeParameterized):
    """
    Provides widgets and callbacks to interact with an AbiPy Flow.
    """

    def __init__(self, flow: Flow, **params):
        NodeParameterized.__init__(self, node=flow, **params)

        self.structures_btn = pnw.Button(name="Show Structures", button_type='primary')
        self.structures_io_checkbox = pnw.CheckBoxGroup(
            name='Input/Output Structure', value=['output'], options=['input', 'output'], inline=True)

        self.wt_selector = WorkTaskSelector(flow)
        self.task_btn = pnw.Button(name="Analyze Task", button_type='primary')

    def get_task_view(self) -> pn.Column:
        wbox = pn.WidgetBox

        return pn.Column(
            wbox("## Select Work and Task",
                 self.wt_selector,
                 self.task_btn,
            ),
            pn.layout.Divider(),
            self.on_task_btn,
            sizing_mode='stretch_width',
        )

    @depends_on_btn_click("task_btn")
    def on_task_btn(self) -> pn.Column:
        """
        Return panel associated to the selected task.
        """
        task = self.wt_selector.task
        return pn.Column(
            f"## {repr(task)}",
            task.get_panel(),
            sizing_mode="stretch_width",
        )

    @depends_on_btn_click("structures_btn")
    def on_structures_btn(self) -> pn.Row:
        what = ""
        if "input" in self.structures_io_checkbox.value: what += "i"
        if "output" in self.structures_io_checkbox.value: what += "o"
        dfs = self.flow.compare_structures(nids=None, # select_nids(flow, options),
                                           what=what,
                                           verbose=self.verbose, with_spglib=False, printout=False,
                                           with_colors=False)

        return pn.Row(dfc(dfs.lattice), sizing_mode="scale_width")

    def get_panel(self, as_dict=False, **kwargs):
        """Return tabs with widgets to interact with the flow."""

        d = super().get_panel(as_dict=True)

        #row = pn.Row(bkw.PreText(text=self.ddb.to_string(verbose=self.verbose), sizing_mode="scale_both"))
        d["Task"] = self.get_task_view()
        #d["Work"] = self.get_work_view()
        #d["Structures"] = pn.Row(pn.Column(self.structures_io_checkbox, self.structures_btn), self.on_structures_btn)
        ###ws = pn.Column(self.ebands_plotter_mode, self.ebands_ksamp_checkbox, self.ebands_df_checkbox, self.ebands_plotter_btn)
        ###d["Ebands"] = pn.Row(ws, self.on_ebands_btn)
        #d["Browse"] = self.get_workdir_view()

        if as_dict: return d

        return self.get_template_from_tabs(d, template=kwargs.get("template", None), closable=False)


class JsPane(pn.pane.HTML):
    """
    Based on: https://discourse.holoviz.org/t/how-to-make-a-dynamic-link-in-panel/2137
    """
    def __init__(self):
        super().__init__(width=0, height=0, margin=0, sizing_mode="fixed")

    def execute_js(self, script: str) -> None:
        script = f'<script type="text/javascript">{script}</script>'
        self.object = script
        self.object = ""


class FlowMultiPageApp():

    def __init__(self, flow: Flow, template, spectator_mode=True, **kwargs):

        if spectator_mode:
            # We are in read-only mode so we have to disable signals to avoid side effects and callbacks
            flow.set_spectator_mode()

        self.flow = flow
        self.template = template

        self.wt_selector = WorkTaskSelector(flow)
        goto_work_btn = pnw.Button(name="Go to Work", button_type='primary')
        goto_work_btn.on_click(self.on_goto_work_bnt)
        goto_task_btn = pnw.Button(name="Go to Task", button_type='primary')
        goto_task_btn.on_click(self.on_goto_task_bnt)
        self.new_tab = pnw.Checkbox(value=True, name="Open in new Tab")
        self.js_panel = JsPane()

        self.sidebar = pn.WidgetBox(
            self.wt_selector,
            self.new_tab,
            pn.layout.Divider(),
            goto_work_btn,
            goto_task_btn,
            self.js_panel,
        )

        def handle_home():
            app = FlowPanel(self.flow).get_panel(template=template)
            if hasattr(app, "sidebar"):
                app.sidebar.append(self.sidebar)
                #app.header.append(self.sidebar)

            return app

        # url --> handler
        self.routes = {
            "/": handle_home,
            r"/w\d+/?": self.handle_wt,
            r"/w\d+/t\d+": self.handle_wt,
        }

    def on_goto_work_bnt(self, event) -> None:
        work = self.wt_selector.work
        url = "/w%d" % work.pos
        code = f"window.open('{url}')" if self.new_tab.value else f"window.location.href='{url}'"
        self.js_panel.execute_js(code)

    def on_goto_task_bnt(self, event) -> None:
        task = self.wt_selector.task
        url = "/w%d/t%d" % (task.pos[0], task.pos[1])
        code = f"window.open('{url}')" if self.new_tab.value else f"window.location.href='{url}'"
        self.js_panel.execute_js(code)

    def handle_wt(self):
        # URL example: /w1/t5/w\d+/t\d+
        #print("in handle_wt with pn.state.app_url:", pn.state.app_url)
        tokens = pn.state.app_url.split("/")
        work_idx = int(tokens[1][1:])
        task_idx = None
        if tokens[2].startswith("t"):
            task_idx = int(tokens[2][1:])

        #print("got request with work_idx:", work_idx, "task_idx:", task_idx)
        #from abipy.panels.core import abipanel
        #abipanel()

        if task_idx is None:
            work = self.flow[work_idx]
            app = work.get_panel(template=self.template, title=repr(work))
        else:
            task = self.flow[work_idx][task_idx]
            app = task.get_panel(template=self.template, title=repr(task))

        if hasattr(app, "sidebar"):
            app.sidebar.append(self.sidebar)
            #app.header.append(self.sidebar)

        #app.title = title

        return app

    def serve(self, **serve_kwargs):
        return pn.serve(self.routes, **serve_kwargs)